from __future__ import annotations

from .resolver import (
    admin_assign_role,
    admin_bulk_assign_role,
    admin_bulk_revoke_role,
    admin_create_permission,
    admin_create_role,
    admin_delete_permission,
    admin_delete_role,
    admin_entities,
    admin_permissions,
    admin_purge_role,
    admin_revoke_role,
    admin_role,
    admin_role_assignments,
    admin_roles,
    admin_update_permission,
    admin_update_role,
    my_roles,
    rbac_scope_entity_combinations,
)
from .types import (
    AssignRoleInput,
    BulkAssignRoleInputGQL,
    BulkAssignRolePayloadGQL,
    BulkRevokeRoleInputGQL,
    BulkRevokeRolePayloadGQL,
    CreatePermissionInput,
    CreateRoleInput,
    EntityConnection,
    EntityFilter,
    EntityOrderBy,
    EntityRefGQL,
    OperationTypeGQL,
    PermissionConnection,
    PermissionFilter,
    PermissionGQL,
    PermissionOrderBy,
    RBACElementTypeGQL,
    RevokeRoleInput,
    RoleAssignmentConnection,
    RoleAssignmentFilter,
    RoleAssignmentGQL,
    RoleConnection,
    RoleFilter,
    RoleGQL,
    RoleOrderBy,
    RoleSourceGQL,
    RoleStatusGQL,
    ScopeEntityCombinationGQL,
    UpdatePermissionInput,
    UpdateRoleInput,
)

__all__ = (
    # Enums
    "RBACElementTypeGQL",
    "OperationTypeGQL",
    "RoleSourceGQL",
    "RoleStatusGQL",
    # Types
    "RoleGQL",
    "PermissionGQL",
    "RoleAssignmentGQL",
    "EntityRefGQL",
    "ScopeEntityCombinationGQL",
    # Filters
    "RoleFilter",
    "PermissionFilter",
    "RoleAssignmentFilter",
    "EntityFilter",
    # OrderBy
    "RoleOrderBy",
    "PermissionOrderBy",
    "EntityOrderBy",
    # Inputs
    "CreateRoleInput",
    "UpdateRoleInput",
    "CreatePermissionInput",
    "UpdatePermissionInput",
    "AssignRoleInput",
    "RevokeRoleInput",
    "BulkAssignRoleInputGQL",
    "BulkRevokeRoleInputGQL",
    # Payloads
    "BulkAssignRolePayloadGQL",
    "BulkRevokeRolePayloadGQL",
    # Connections
    "RoleConnection",
    "PermissionConnection",
    "RoleAssignmentConnection",
    "EntityConnection",
    # Query resolvers
    "admin_role",
    "admin_roles",
    "admin_permissions",
    "admin_role_assignments",
    "admin_entities",
    "my_roles",
    "rbac_scope_entity_combinations",
    # Mutation resolvers
    "admin_create_role",
    "admin_update_role",
    "admin_delete_role",
    "admin_purge_role",
    "admin_create_permission",
    "admin_update_permission",
    "admin_delete_permission",
    "admin_assign_role",
    "admin_revoke_role",
    "admin_bulk_assign_role",
    "admin_bulk_revoke_role",
)



from .bgtask_registry import BgtaskRegistryDependency, BgtaskRegistryInput
from .composer import ProcessingComposer, ProcessingInput, ProcessingResources
from .event_dispatcher import EventDispatcherDependency, EventDispatcherInput
from .processors import ProcessorsDependency, ProcessorsProviderInput

__all__ = [
    "BgtaskRegistryDependency",
    "BgtaskRegistryInput",
    "EventDispatcherDependency",
    "EventDispatcherInput",
    "ProcessingComposer",
    "ProcessingInput",
    "ProcessingResources",
    "ProcessorsDependency",
    "ProcessorsProviderInput",
]



from .actions import (
    FetchManagerStatusAction,
    FetchManagerStatusActionResult,
    GetAnnouncementAction,
    GetAnnouncementActionResult,
    GetDbCxnStatusAction,
    GetDbCxnStatusActionResult,
    PerformSchedulerOpsAction,
    PerformSchedulerOpsActionResult,
    UpdateAnnouncementAction,
    UpdateAnnouncementActionResult,
    UpdateManagerStatusAction,
    UpdateManagerStatusActionResult,
)

__all__ = (
    "FetchManagerStatusAction",
    "FetchManagerStatusActionResult",
    "GetAnnouncementAction",
    "GetAnnouncementActionResult",
    "GetDbCxnStatusAction",
    "GetDbCxnStatusActionResult",
    "PerformSchedulerOpsAction",
    "PerformSchedulerOpsActionResult",
    "UpdateAnnouncementAction",
    "UpdateAnnouncementActionResult",
    "UpdateManagerStatusAction",
    "UpdateManagerStatusActionResult",
)



"""Auth handler class using constructor dependency injection.

All handlers use the new ApiHandler pattern: typed parameters
(``BodyParam``, ``QueryParam``, ``UserContext``, ``RequestCtx``) are
automatically extracted by ``_wrap_api_handler`` and responses are
returned as ``APIResponse`` objects.
"""

from __future__ import annotations

import logging
from http import HTTPStatus
from typing import Final

from aiohttp import web

from ai.backend.common.api_handlers import APIResponse, BodyParam, QueryParam
from ai.backend.common.dto.manager.auth.request import (
    AuthorizeRequest,
    GetRoleRequest,
    SignoutRequest,
    SignupRequest,
    UpdateFullNameRequest,
    UpdatePasswordNoAuthRequest,
    UpdatePasswordRequest,
    UploadSSHKeypairRequest,
    VerifyAuthRequest,
)
from ai.backend.common.dto.manager.auth.response import (
    AuthorizeResponse,
    GetRoleResponse,
    GetSSHKeypairResponse,
    SignoutResponse,
    SignupResponse,
    SSHKeypairResponse,
    UpdateFullNameResponse,
    UpdatePasswordNoAuthResponse,
    UpdatePasswordResponse,
    VerifyAuthResponse,
)
from ai.backend.common.dto.manager.auth.types import (
    AuthResponseType,
    AuthSuccessResponse,
    AuthTokenType,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.dto.context import RequestCtx, UserContext
from ai.backend.manager.errors.auth import AuthorizationFailed
from ai.backend.manager.services.auth.actions.authorize import AuthorizeAction
from ai.backend.manager.services.auth.actions.generate_ssh_keypair import GenerateSSHKeypairAction
from ai.backend.manager.services.auth.actions.get_role import GetRoleAction
from ai.backend.manager.services.auth.actions.get_ssh_keypair import GetSSHKeypairAction
from ai.backend.manager.services.auth.actions.signout import SignoutAction
from ai.backend.manager.services.auth.actions.signup import SignupAction
from ai.backend.manager.services.auth.actions.update_full_name import UpdateFullNameAction
from ai.backend.manager.services.auth.actions.update_password import UpdatePasswordAction
from ai.backend.manager.services.auth.actions.update_password_no_auth import (
    UpdatePasswordNoAuthAction,
)
from ai.backend.manager.services.auth.actions.upload_ssh_keypair import UploadSSHKeypairAction
from ai.backend.manager.services.auth.processors import AuthProcessors

log: Final = BraceStyleAdapter(logging.getLogger(__spec__.name))


class AuthHandler:
    """Auth API handler with constructor-injected dependencies."""

    def __init__(self, *, auth: AuthProcessors) -> None:
        self._auth = auth

    # ------------------------------------------------------------------
    # test_get (GET /auth, /auth/test)
    # ------------------------------------------------------------------

    async def test_get(self, ctx: UserContext) -> APIResponse:
        log.info("AUTH.TEST(ak:{})", ctx.access_key)
        resp = VerifyAuthResponse(authorized="yes", echo="")
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # test_post (POST /auth, /auth/test)
    # ------------------------------------------------------------------

    async def test_post(self, body: BodyParam[VerifyAuthRequest], ctx: UserContext) -> APIResponse:
        log.info("AUTH.TEST(ak:{})", ctx.access_key)
        params = body.parsed
        resp = VerifyAuthResponse(authorized="yes", echo=params.echo)
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # get_role (GET /auth/role)
    # ------------------------------------------------------------------

    async def get_role(self, query: QueryParam[GetRoleRequest], ctx: UserContext) -> APIResponse:
        params = query.parsed
        log.info(
            "AUTH.ROLES(ak:{}, d:{}, g:{})",
            ctx.access_key,
            ctx.user_domain,
            params.group,
        )
        action = GetRoleAction(
            user_id=ctx.user_uuid,
            group_id=params.group,
            is_superadmin=ctx.is_superadmin,
            is_admin=ctx.is_admin,
        )
        result = await self._auth.get_role.wait_for_complete(action)
        resp = GetRoleResponse(
            global_role=result.global_role,
            domain_role=result.domain_role,
            group_role=result.group_role,
        )
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # authorize (POST /auth/authorize)
    # ------------------------------------------------------------------

    async def authorize(
        self, body: BodyParam[AuthorizeRequest], ctx: RequestCtx
    ) -> APIResponse | web.StreamResponse:
        params = body.parsed
        log.info(
            "AUTH.AUTHORIZE(d:{}, u:{}, passwd:****, type:{})",
            params.domain,
            params.username,
            params.type,
        )
        action = AuthorizeAction(
            request=ctx.request,
            type=AuthTokenType(params.type),
            domain_name=params.domain,
            email=params.username,
            password=params.password,
            stoken=params.stoken,
        )
        result = await self._auth.authorize.wait_for_complete(action)

        if result.stream_response is not None:
            return result.stream_response

        if result.authorization_result is None:
            raise AuthorizationFailed("Authorization result is missing")
        auth_result = result.authorization_result
        data = AuthSuccessResponse(
            response_type=AuthResponseType.SUCCESS,
            access_key=auth_result.access_key,
            secret_key=auth_result.secret_key,
            role=auth_result.role,
            status=auth_result.status,
        )
        resp = AuthorizeResponse(data=data)
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # signup (POST /auth/signup)
    # ------------------------------------------------------------------

    async def signup(self, body: BodyParam[SignupRequest], ctx: RequestCtx) -> APIResponse:
        params = body.parsed
        log.info("AUTH.SIGNUP(d:{}, email:{}, passwd:****)", params.domain, params.email)
        action = SignupAction(
            request=ctx.request,
            domain_name=params.domain,
            email=params.email,
            password=params.password,
            username=params.username,
            full_name=params.full_name,
            description=params.description,
        )
        result = await self._auth.signup.wait_for_complete(action)
        resp = SignupResponse(
            access_key=result.access_key,
            secret_key=result.secret_key,
        )
        return APIResponse.build(HTTPStatus.CREATED, resp)

    # ------------------------------------------------------------------
    # signout (POST /auth/signout)
    # ------------------------------------------------------------------

    async def signout(self, body: BodyParam[SignoutRequest], ctx: UserContext) -> APIResponse:
        params = body.parsed
        log.info("AUTH.SIGNOUT(d:{}, email:{})", ctx.user_domain, params.email)
        await self._auth.signout.wait_for_complete(
            SignoutAction(
                user_id=ctx.user_uuid,
                domain_name=ctx.user_domain,
                requester_email=ctx.user_email,
                email=params.email,
                password=params.password,
            )
        )
        return APIResponse.build(HTTPStatus.OK, SignoutResponse())

    # ------------------------------------------------------------------
    # update_full_name (POST /auth/update-full-name)
    # ------------------------------------------------------------------

    async def update_full_name(
        self, body: BodyParam[UpdateFullNameRequest], ctx: UserContext
    ) -> APIResponse:
        params = body.parsed
        log.info("AUTH.UPDATE_FULL_NAME(d:{}, email:{})", ctx.user_domain, ctx.user_email)
        await self._auth.update_full_name.wait_for_complete(
            UpdateFullNameAction(
                user_id=str(ctx.user_uuid),
                full_name=params.full_name,
                domain_name=ctx.user_domain,
                email=ctx.user_email,
            )
        )
        return APIResponse.build(HTTPStatus.OK, UpdateFullNameResponse())

    # ------------------------------------------------------------------
    # update_password (POST /auth/update-password)
    # ------------------------------------------------------------------

    async def update_password(
        self,
        body: BodyParam[UpdatePasswordRequest],
        ctx: UserContext,
        req: RequestCtx,
    ) -> APIResponse:
        params = body.parsed
        log.info("AUTH.UPDATE_PASSWORD(d:{}, email:{})", ctx.user_domain, ctx.user_email)
        action = UpdatePasswordAction(
            request=req.request,
            user_id=ctx.user_uuid,
            domain_name=ctx.user_domain,
            email=ctx.user_email,
            old_password=params.old_password,
            new_password=params.new_password,
            new_password_confirm=params.new_password2,
        )
        result = await self._auth.update_password.wait_for_complete(action)
        if not result.success:
            resp = UpdatePasswordResponse(error_msg="new password mismatch")
            return APIResponse.build(HTTPStatus.BAD_REQUEST, resp)
        return APIResponse.build(HTTPStatus.OK, UpdatePasswordResponse())

    # ------------------------------------------------------------------
    # update_password_no_auth (POST /auth/update-password-no-auth)
    # ------------------------------------------------------------------

    async def update_password_no_auth(
        self, body: BodyParam[UpdatePasswordNoAuthRequest], ctx: RequestCtx
    ) -> APIResponse:
        params = body.parsed
        log.info(
            "AUTH.UPDATE_PASSWORD_NO_AUTH(d:{}, u:{}, passwd:****)",
            params.domain,
            params.username,
        )
        action = UpdatePasswordNoAuthAction(
            request=ctx.request,
            domain_name=params.domain,
            email=params.username,
            current_password=params.current_password,
            new_password=params.new_password,
        )
        result = await self._auth.update_password_no_auth.wait_for_complete(action)
        resp = UpdatePasswordNoAuthResponse(
            password_changed_at=result.password_changed_at.isoformat(),
        )
        return APIResponse.build(HTTPStatus.CREATED, resp)

    # ------------------------------------------------------------------
    # get_ssh_keypair (GET /auth/ssh-keypair)
    # ------------------------------------------------------------------

    async def get_ssh_keypair(self, ctx: UserContext) -> APIResponse:
        log.info("AUTH.GET_SSH_KEYPAIR(d:{}, ak:{})", ctx.user_domain, ctx.access_key)
        result = await self._auth.get_ssh_keypair.wait_for_complete(
            GetSSHKeypairAction(
                user_id=ctx.user_uuid,
                access_key=ctx.access_key,
            )
        )
        resp = GetSSHKeypairResponse(ssh_public_key=result.public_key)
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # generate_ssh_keypair (PATCH /auth/ssh-keypair)
    # ------------------------------------------------------------------

    async def generate_ssh_keypair(self, ctx: UserContext) -> APIResponse:
        log.info("AUTH.REFRESH_SSH_KEYPAIR(d:{}, ak:{})", ctx.user_domain, ctx.access_key)
        result = await self._auth.generate_ssh_keypair.wait_for_complete(
            GenerateSSHKeypairAction(
                user_id=ctx.user_uuid,
                access_key=ctx.access_key,
            )
        )
        resp = SSHKeypairResponse(
            ssh_public_key=result.ssh_keypair.ssh_public_key,
            ssh_private_key=result.ssh_keypair.ssh_private_key,
        )
        return APIResponse.build(HTTPStatus.OK, resp)

    # ------------------------------------------------------------------
    # upload_ssh_keypair (POST /auth/ssh-keypair)
    # ------------------------------------------------------------------

    async def upload_ssh_keypair(
        self, body: BodyParam[UploadSSHKeypairRequest], ctx: UserContext
    ) -> APIResponse:
        params = body.parsed
        pubkey = f"{params.pubkey.rstrip()}\n"
        privkey = f"{params.privkey.rstrip()}\n"
        log.info("AUTH.SAVE_SSH_KEYPAIR(d:{}, ak:{})", ctx.user_domain, ctx.access_key)
        result = await self._auth.upload_ssh_keypair.wait_for_complete(
            UploadSSHKeypairAction(
                user_id=ctx.user_uuid,
                public_key=pubkey,
                private_key=privkey,
                access_key=ctx.access_key,
            )
        )
        resp = SSHKeypairResponse(
            ssh_public_key=result.ssh_keypair.ssh_public_key,
            ssh_private_key=result.ssh_keypair.ssh_private_key,
        )
        return APIResponse.build(HTTPStatus.OK, resp)



from __future__ import annotations

import asyncio
import importlib.resources
import logging
from collections.abc import Callable, Mapping, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any, Final

import aiohttp_cors
from aiohttp import web

from ai.backend.common.metrics.metric import CommonMetricRegistry
from ai.backend.common.metrics.profiler import Profiler, PyroscopeArgs
from ai.backend.common.types import AgentSelectionStrategy
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager import __version__
from ai.backend.manager.config.bootstrap import BootstrapConfig
from ai.backend.manager.errors.common import (
    GenericBadRequest,
    InternalServerError,
    ServerMisconfiguredError,
)

from .middleware import build_api_metric_middleware, request_id_middleware
from .routing import RouteRegistry

if TYPE_CHECKING:
    from .types import CORSOptions, WebRequestHandler

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

VALID_VERSIONS: Final = frozenset([
    "v4.20181215",
    "v4.20190115",
    "v4.20190315",
    "v4.20190615",
    "v5.20191215",
    "v6.20200815",
    "v6.20220315",
    "v6.20220615",
    "v6.20221201",
    "v6.20230315",
    "v7.20230615",
    "v8.20240315",
    "v8.20240915",
    "v9.20250722",
])
LATEST_REV_DATES: Final = {
    1: "20160915",
    2: "20170915",
    3: "20181215",
    4: "20190615",
    5: "20191215",
    6: "20230315",
    7: "20230615",
    8: "20240915",
    9: "20250722",
}
LATEST_API_VERSION: Final = (
    f"v{max(LATEST_REV_DATES.keys())}.{LATEST_REV_DATES[max(LATEST_REV_DATES.keys())]}"
)


async def hello(_request: web.Request) -> web.Response:
    """
    Returns the API version number.
    """
    return web.json_response({
        "version": LATEST_API_VERSION,
        "manager": __version__,
    })


async def on_prepare(_request: web.Request, response: web.StreamResponse) -> None:
    response.headers["Server"] = "BackendAI"


@web.middleware
async def api_middleware(request: web.Request, handler: WebRequestHandler) -> web.StreamResponse:
    _handler = handler
    method_override = request.headers.get("X-Method-Override", None)
    if method_override:
        request = request.clone(method=method_override)
        new_match_info = await request.app.router.resolve(request)
        if new_match_info is None:
            raise InternalServerError("No matching method handler found")
        _handler = new_match_info.handler
        request._match_info = new_match_info
    ex = request.match_info.http_exception
    if ex is not None:
        raise ex
    new_api_version = request.headers.get("X-BackendAI-Version")
    legacy_api_version = request.headers.get("X-Sorna-Version")
    api_version = new_api_version or legacy_api_version
    try:
        if api_version is None:
            path_major_version = int(request.match_info.get("version", 5))
            revision_date = LATEST_REV_DATES[path_major_version]
            request["api_version"] = (path_major_version, revision_date)
        elif api_version in VALID_VERSIONS:
            hdr_major_version, revision_date = api_version.split(".", maxsplit=1)
            request["api_version"] = (int(hdr_major_version[1:]), revision_date)
        else:
            return GenericBadRequest("Unsupported API version.")
    except (ValueError, KeyError):
        return GenericBadRequest("Unsupported API version.")
    return await _handler(request)


def _mount_registry_tree(
    root_app: web.Application,
    root_registry: RouteRegistry,
    pidx: int = 0,
) -> None:
    """Flatten the registry tree and mount all subapps on *root_app*."""

    async def _bridge_root_app(subapp: web.Application) -> None:
        subapp["_root_app"] = root_app

    for prefix, app, _reg in root_registry.collect_apps():
        if pidx == 0:
            log.info("Loading module: {}", prefix)
        app["_registry_prefix"] = prefix
        app.on_startup.insert(0, _bridge_root_app)
        root_app.add_subapp("/" + prefix, app)


def mount_registries(
    root_app: web.Application,
    registries: Sequence[RouteRegistry],
    *,
    cors_options: CORSOptions | None = None,
) -> None:
    """Mount pre-built registries on *root_app*.

    Public API used by ``tests/component/conftest.py`` to mount only
    the modules needed for a particular test.
    """
    root_registry = RouteRegistry.create("", cors_options or {})
    for reg in registries:
        root_registry.add_subregistry(reg)
    _mount_registry_tree(root_app, root_registry)

    # Install ratelimit middleware on root app if the module is present
    rlim_reg = root_registry.find_subregistry("ratelimit")
    if rlim_reg is not None and rlim_reg.rlim_middleware is not None:
        root_app.middlewares.append(rlim_reg.rlim_middleware)


def build_root_app(
    pidx: int,
    bootstrap_config: BootstrapConfig,
    *,
    scheduler_opts: Mapping[str, Any] | None = None,
    loop_error_handler: Callable[..., Any] | None = None,
) -> web.Application:
    if bootstrap_config.pyroscope.enabled:
        if (
            not bootstrap_config.pyroscope.app_name
            or not bootstrap_config.pyroscope.server_addr
            or not bootstrap_config.pyroscope.sample_rate
        ):
            raise ValueError("Pyroscope configuration is incomplete.")

        Profiler(
            pyroscope_args=PyroscopeArgs(
                enabled=bootstrap_config.pyroscope.enabled,
                application_name=bootstrap_config.pyroscope.app_name,
                server_address=bootstrap_config.pyroscope.server_addr,
                sample_rate=bootstrap_config.pyroscope.sample_rate,
            )
        )

    metrics = CommonMetricRegistry.instance()
    cors_options = {
        "*": aiohttp_cors.ResourceOptions(  # type: ignore[no-untyped-call]
            allow_credentials=False, expose_headers="*", allow_headers="*"
        ),
    }
    app = web.Application(
        middlewares=[
            request_id_middleware,
            # exception_middleware and auth_middleware are inserted later
            # in server_main() after dependencies are available.
            api_middleware,
            build_api_metric_middleware(metrics.api),
        ]
    )
    if loop_error_handler is not None:
        loop = asyncio.get_running_loop()
        loop.set_exception_handler(loop_error_handler)
    global_exception_handler = loop_error_handler

    # If the request path starts with the following route, the auth_middleware is bypassed.
    # In this case, all authentication flags are turned off.
    # Used in special cases where the request headers cannot be modified.
    app["auth_middleware_allowlist"] = [
        "/container-registries/webhook",
    ]

    app["_pidx"] = pidx
    app["_cors_options"] = cors_options
    app["_metrics"] = metrics
    default_scheduler_opts = {
        "limit": 2048,
        "close_timeout": 30,
        "exception_handler": global_exception_handler,
        "agent_selection_strategy": AgentSelectionStrategy.DISPERSED,
    }
    app["scheduler_opts"] = {
        **default_scheduler_opts,
        **(scheduler_opts if scheduler_opts is not None else {}),
    }
    app.on_response_prepare.append(on_prepare)

    cors = aiohttp_cors.setup(app, defaults=cors_options)
    # should be done in create_app() in other modules.
    cors.add(app.router.add_route("GET", r"", hello))
    cors.add(app.router.add_route("GET", r"/", hello))

    vendor_path = importlib.resources.files("ai.backend.manager.vendor")
    if not isinstance(vendor_path, Path):
        raise ServerMisconfiguredError("vendor_path must be a Path instance")
    app.router.add_static("/static/vendor", path=vendor_path, name="static")
    return app



from __future__ import annotations

from dataclasses import dataclass
from typing import override
from uuid import UUID

from ai.backend.common.data.permission.types import RBACElementType
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.permission.types import RBACElementRef

from .base import NotificationChannelSingleEntityAction, NotificationChannelSingleEntityActionResult


@dataclass
class ValidateChannelAction(NotificationChannelSingleEntityAction):
    """Action to validate a notification channel (webhook sending test)."""

    channel_id: UUID
    test_message: str

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE

    @override
    def target_entity_id(self) -> str:
        return str(self.channel_id)

    @override
    def target_element(self) -> RBACElementRef:
        return RBACElementRef(RBACElementType.NOTIFICATION_CHANNEL, str(self.channel_id))


@dataclass
class ValidateChannelActionResult(NotificationChannelSingleEntityActionResult):
    """Result of validating a notification channel."""

    @override
    def target_entity_id(self) -> str:
        return ""



"""Prometheus query preset GQL mutation resolvers."""

from __future__ import annotations

from uuid import UUID

import strawberry
from strawberry import ID, Info

from ai.backend.manager.api.gql.prometheus_query_preset.types import (
    CreateQueryDefinitionInput,
    CreateQueryDefinitionPayload,
    DeleteQueryDefinitionPayload,
    ModifyQueryDefinitionInput,
    ModifyQueryDefinitionPayload,
    QueryDefinitionGQL,
)
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.api.gql.utils import check_admin_only
from ai.backend.manager.services.prometheus_query_preset.actions import (
    CreatePresetAction,
    DeletePresetAction,
    ModifyPresetAction,
)


@strawberry.mutation(description="Added in 26.3.0. Create a new query definition (admin only).")  # type: ignore[misc]
async def admin_create_prometheus_query_preset(
    info: Info[StrawberryGQLContext],
    input: CreateQueryDefinitionInput,
) -> CreateQueryDefinitionPayload:
    check_admin_only()
    processors = info.context.processors

    action_result = await processors.prometheus_query_preset.create_preset.wait_for_complete(
        CreatePresetAction(creator=input.to_creator())
    )

    return CreateQueryDefinitionPayload(preset=QueryDefinitionGQL.from_data(action_result.preset))


@strawberry.mutation(
    description="Added in 26.3.0. Modify an existing query definition (admin only)."
)  # type: ignore[misc]
async def admin_modify_prometheus_query_preset(
    info: Info[StrawberryGQLContext],
    id: ID,
    input: ModifyQueryDefinitionInput,
) -> ModifyQueryDefinitionPayload:
    check_admin_only()
    processors = info.context.processors

    preset_id = UUID(id)
    action_result = await processors.prometheus_query_preset.modify_preset.wait_for_complete(
        ModifyPresetAction(preset_id=preset_id, updater=input.to_updater(preset_id))
    )

    return ModifyQueryDefinitionPayload(preset=QueryDefinitionGQL.from_data(action_result.preset))


@strawberry.mutation(description="Added in 26.3.0. Delete a query definition (admin only).")  # type: ignore[misc]
async def admin_delete_prometheus_query_preset(
    info: Info[StrawberryGQLContext],
    id: ID,
) -> DeleteQueryDefinitionPayload:
    check_admin_only()
    processors = info.context.processors

    await processors.prometheus_query_preset.delete_preset.wait_for_complete(
        DeletePresetAction(preset_id=UUID(id))
    )

    return DeleteQueryDefinitionPayload(id=id)



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.resource_slot.types import AgentResourceData

from .base import ResourceSlotAction


@dataclass
class GetAgentResourceBySlotAction(ResourceSlotAction):
    agent_id: str
    slot_name: str

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.AGENT_RESOURCE

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET

    @override
    def entity_id(self) -> str | None:
        return f"{self.agent_id}:{self.slot_name}"


@dataclass
class GetAgentResourceBySlotResult(BaseActionResult):
    item: AgentResourceData

    @override
    def entity_id(self) -> str | None:
        return None



"""Object storage module registrar."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.manager.api.rest.middleware.auth import auth_required
from ai.backend.manager.api.rest.routing import RouteRegistry

from .handler import ObjectStorageHandler

if TYPE_CHECKING:
    from ai.backend.manager.api.rest.types import RouteDeps


def register_object_storage_routes(
    handler: ObjectStorageHandler, route_deps: RouteDeps
) -> RouteRegistry:
    """Build the object storage sub-application."""

    reg = RouteRegistry.create("object-storages", route_deps.cors_options)

    reg.add(
        "POST",
        "/presigned/upload",
        handler.get_presigned_upload_url,
        middlewares=[auth_required],
    )
    reg.add(
        "POST",
        "/presigned/download",
        handler.get_presigned_download_url,
        middlewares=[auth_required],
    )
    # TODO: deprecate these APIs, and use /storage-namespaces instead
    reg.add(
        "GET",
        "/buckets",
        handler.get_all_buckets,
        middlewares=[auth_required],
    )
    reg.add(
        "GET",
        "/{storage_id}/buckets",
        handler.get_buckets,
        middlewares=[auth_required],
    )
    reg.add(
        "GET",
        "/",
        handler.list_object_storages,
        middlewares=[auth_required],
    )
    return reg



from .actions import (
    CreateClusterTemplateAction,
    CreateClusterTemplateActionResult,
    CreatedTaskTemplateItem,
    CreateTaskTemplateAction,
    CreateTaskTemplateActionResult,
    DeleteClusterTemplateAction,
    DeleteClusterTemplateActionResult,
    DeleteTaskTemplateAction,
    DeleteTaskTemplateActionResult,
    GetClusterTemplateAction,
    GetClusterTemplateActionResult,
    GetTaskTemplateAction,
    GetTaskTemplateActionResult,
    ListClusterTemplatesAction,
    ListClusterTemplatesActionResult,
    ListTaskTemplatesAction,
    ListTaskTemplatesActionResult,
    TaskTemplateItemInput,
    UpdateClusterTemplateAction,
    UpdateClusterTemplateActionResult,
    UpdateTaskTemplateAction,
    UpdateTaskTemplateActionResult,
)

__all__ = (
    "CreateClusterTemplateAction",
    "CreateClusterTemplateActionResult",
    "CreateTaskTemplateAction",
    "CreateTaskTemplateActionResult",
    "CreatedTaskTemplateItem",
    "DeleteClusterTemplateAction",
    "DeleteClusterTemplateActionResult",
    "DeleteTaskTemplateAction",
    "DeleteTaskTemplateActionResult",
    "GetClusterTemplateAction",
    "GetClusterTemplateActionResult",
    "GetTaskTemplateAction",
    "GetTaskTemplateActionResult",
    "ListClusterTemplatesAction",
    "ListClusterTemplatesActionResult",
    "ListTaskTemplatesAction",
    "ListTaskTemplatesActionResult",
    "TaskTemplateItemInput",
    "UpdateClusterTemplateAction",
    "UpdateClusterTemplateActionResult",
    "UpdateTaskTemplateAction",
    "UpdateTaskTemplateActionResult",
)



"""Resource Slot Repository with Resilience policies."""

from __future__ import annotations

import uuid
from decimal import Decimal
from typing import TYPE_CHECKING

from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience import (
    MetricArgs,
    MetricPolicy,
    Resilience,
    RetryArgs,
    RetryPolicy,
)
from ai.backend.common.resilience.policies.retry import BackoffStrategy
from ai.backend.manager.data.resource_slot.types import (
    AgentResourceDrift,
    AgentResourceSearchResult,
    ResourceAllocationSearchResult,
    ResourceOccupancy,
    ResourceSlotTypeSearchResult,
)
from ai.backend.manager.models.resource_slot import (
    AgentResourceRow,
    ResourceAllocationRow,
    ResourceSlotTypeRow,
)
from ai.backend.manager.repositories.base import BatchQuerier

from .db_source import ResourceSlotDBSource

if TYPE_CHECKING:
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

__all__ = ("ResourceSlotRepository",)


resource_slot_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.RESOURCE_SLOT_REPOSITORY)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
            )
        ),
    ]
)


class ResourceSlotRepository:
    """Repository for normalized resource slot data access with resilience policies.

    Manages three tables:
    - resource_slot_types: Registry of known slot types
    - agent_resources: Per-agent, per-slot capacity and usage
    - resource_allocations: Per-kernel, per-slot resource allocation
    """

    _db_source: ResourceSlotDBSource

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db_source = ResourceSlotDBSource(db)

    # ==================== resource_slot_types ====================

    @resource_slot_repository_resilience.apply()
    async def all_slot_types(self) -> list[ResourceSlotTypeRow]:
        """List all registered resource slot types."""
        return await self._db_source.all_slot_types()

    @resource_slot_repository_resilience.apply()
    async def get_slot_type(self, slot_name: str) -> ResourceSlotTypeRow:
        """Get a specific resource slot type by name."""
        return await self._db_source.get_slot_type(slot_name)

    @resource_slot_repository_resilience.apply()
    async def search_slot_types(self, querier: BatchQuerier) -> ResourceSlotTypeSearchResult:
        """Paginated search across resource slot types."""
        return await self._db_source.search_slot_types(querier)

    # ==================== agent_resources ====================

    @resource_slot_repository_resilience.apply()
    async def get_agent_resources(self, agent_id: str) -> list[AgentResourceRow]:
        """Get all slot capacity/usage rows for a given agent."""
        return await self._db_source.get_agent_resources(agent_id)

    @resource_slot_repository_resilience.apply()
    async def get_agent_resource_by_slot(self, agent_id: str, slot_name: str) -> AgentResourceRow:
        """Get a single slot row for one agent+slot combination."""
        return await self._db_source.get_agent_resource_by_slot(agent_id, slot_name)

    @resource_slot_repository_resilience.apply()
    async def search_agent_resources(self, querier: BatchQuerier) -> AgentResourceSearchResult:
        return await self._db_source.search_agent_resources(querier)

    # ==================== resource_allocations ====================

    @resource_slot_repository_resilience.apply()
    async def get_kernel_allocations(self, kernel_id: uuid.UUID) -> list[ResourceAllocationRow]:
        """Get all per-slot allocation rows for a given kernel."""
        return await self._db_source.get_kernel_allocations(kernel_id)

    @resource_slot_repository_resilience.apply()
    async def get_kernel_allocation_by_slot(
        self, kernel_id: uuid.UUID, slot_name: str
    ) -> ResourceAllocationRow:
        """Get a single allocation row for one kernel+slot combination."""
        return await self._db_source.get_kernel_allocation_by_slot(kernel_id, slot_name)

    @resource_slot_repository_resilience.apply()
    async def search_resource_allocations(
        self, querier: BatchQuerier
    ) -> ResourceAllocationSearchResult:
        return await self._db_source.search_resource_allocations(querier)

    # ==================== Aggregation ====================

    @resource_slot_repository_resilience.apply()
    async def compute_actual_agent_resource_usage(
        self,
    ) -> dict[tuple[str, str], Decimal]:
        """Compute actual per-agent per-slot resource usage from active allocations."""
        return await self._db_source.compute_actual_agent_resource_usage()

    @resource_slot_repository_resilience.apply()
    async def reconcile_agent_resources(self) -> list[AgentResourceDrift]:
        """Compare agent_resources.used against actual allocations and correct drift."""
        return await self._db_source.reconcile_agent_resources()

    @resource_slot_repository_resilience.apply()
    async def get_domain_resource_overview(self, domain_name: str) -> ResourceOccupancy:
        """Get aggregated active resource occupancy for a domain."""
        return await self._db_source.aggregate_occupied_by_domain(domain_name)

    @resource_slot_repository_resilience.apply()
    async def get_project_resource_overview(self, project_id: uuid.UUID) -> ResourceOccupancy:
        """Get aggregated active resource occupancy for a project (group)."""
        return await self._db_source.aggregate_occupied_by_project(project_id)



"""
Adapters to convert compute session DTOs to repository BatchQuerier objects.
Handles conversion of filter, order, and pagination parameters.
Also provides data-to-DTO conversion functions.
"""

from __future__ import annotations

from collections import defaultdict
from typing import Any
from uuid import UUID

from ai.backend.common.dto.manager.compute_session import (
    ComputeSessionDTO,
    ComputeSessionFilter,
    ComputeSessionOrder,
    ComputeSessionOrderField,
    ContainerDTO,
    OrderDirection,
    SearchComputeSessionsRequest,
)
from ai.backend.common.types import SessionId
from ai.backend.manager.api.rest.adapter import BaseFilterAdapter
from ai.backend.manager.data.kernel.types import KernelInfo
from ai.backend.manager.data.session.types import SessionData, SessionStatus
from ai.backend.manager.repositories.base import (
    BatchQuerier,
    NoPagination,
    OffsetPagination,
    QueryCondition,
    QueryOrder,
)
from ai.backend.manager.repositories.scheduler.options import (
    KernelConditions,
    SessionConditions,
    SessionOrders,
)


class ComputeSessionsAdapter(BaseFilterAdapter):
    """Adapter for converting compute session requests to repository queries."""

    def build_session_querier(self, request: SearchComputeSessionsRequest) -> BatchQuerier:
        """Build a BatchQuerier for compute sessions from search request."""
        conditions = self._convert_session_filter(request.filter) if request.filter else []
        orders = [self._convert_session_order(o) for o in request.order] if request.order else []
        pagination = OffsetPagination(limit=request.limit, offset=request.offset)

        return BatchQuerier(conditions=conditions, orders=orders, pagination=pagination)

    def build_kernel_querier_for_sessions(self, session_ids: list[SessionId]) -> BatchQuerier:
        """Build a BatchQuerier for kernels belonging to the given sessions."""
        conditions: list[QueryCondition] = [KernelConditions.by_session_ids(session_ids)]
        return BatchQuerier(conditions=conditions, orders=[], pagination=NoPagination())

    def group_kernels_by_session(self, kernels: list[KernelInfo]) -> dict[UUID, list[KernelInfo]]:
        """Group kernel info list by session ID."""
        grouped: dict[UUID, list[KernelInfo]] = defaultdict(list)
        for kernel in kernels:
            session_id = UUID(kernel.session.session_id)
            grouped[session_id].append(kernel)
        return grouped

    def convert_session_to_dto(
        self, session: SessionData, kernels: list[KernelInfo] | None = None
    ) -> ComputeSessionDTO:
        """Convert SessionData + kernels to ComputeSessionDTO."""
        containers = [self._convert_kernel_to_container(k) for k in kernels] if kernels else []

        resource_slots: dict[str, Any] | None = None
        if session.requested_slots is not None:
            resource_slots = dict(session.requested_slots)

        occupied_slots: dict[str, Any] | None = None
        if session.occupying_slots is not None:
            occupied_slots = dict(session.occupying_slots)

        return ComputeSessionDTO(
            id=session.id,
            name=session.name,
            type=session.session_type.value,
            status=session.status.value,
            image=session.images,
            scaling_group=session.scaling_group_name,
            resource_slots=resource_slots,
            occupied_slots=occupied_slots,
            created_at=session.created_at,
            terminated_at=session.terminated_at,
            starts_at=session.starts_at,
            containers=containers,
        )

    def _convert_kernel_to_container(self, kernel: KernelInfo) -> ContainerDTO:
        """Convert KernelInfo to ContainerDTO."""
        resource_usage: dict[str, Any] | None = None
        if kernel.metrics.last_stat is not None:
            resource_usage = kernel.metrics.last_stat

        return ContainerDTO(
            id=kernel.id,
            agent_id=kernel.resource.agent,
            status=kernel.lifecycle.status.value,
            resource_usage=resource_usage,
        )

    def _convert_session_filter(self, filter: ComputeSessionFilter) -> list[QueryCondition]:
        """Convert session filter to list of query conditions."""
        conditions: list[QueryCondition] = []

        if filter.status is not None and len(filter.status) > 0:
            statuses = [SessionStatus(s) for s in filter.status]
            conditions.append(SessionConditions.by_statuses(statuses))

        if filter.name is not None:
            condition = self.convert_string_filter(
                filter.name,
                contains_factory=SessionConditions.by_name_contains,
                equals_factory=SessionConditions.by_name_equals,
                starts_with_factory=SessionConditions.by_name_starts_with,
                ends_with_factory=SessionConditions.by_name_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        if filter.access_key is not None:
            condition = self.convert_string_filter(
                filter.access_key,
                contains_factory=SessionConditions.by_access_key_contains,
                equals_factory=SessionConditions.by_access_key_equals,
                starts_with_factory=SessionConditions.by_access_key_starts_with,
                ends_with_factory=SessionConditions.by_access_key_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        if filter.domain_name is not None:
            condition = self.convert_string_filter(
                filter.domain_name,
                contains_factory=SessionConditions.by_domain_name_contains,
                equals_factory=SessionConditions.by_domain_name_equals,
                starts_with_factory=SessionConditions.by_domain_name_starts_with,
                ends_with_factory=SessionConditions.by_domain_name_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        if filter.scaling_group_name is not None:
            condition = self.convert_string_filter(
                filter.scaling_group_name,
                contains_factory=SessionConditions.by_scaling_group_contains,
                equals_factory=SessionConditions.by_scaling_group_equals,
                starts_with_factory=SessionConditions.by_scaling_group_starts_with,
                ends_with_factory=SessionConditions.by_scaling_group_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        return conditions

    def _convert_session_order(self, order: ComputeSessionOrder) -> QueryOrder:
        """Convert session order specification to query order."""
        ascending = order.direction == OrderDirection.ASC

        match order.field:
            case ComputeSessionOrderField.CREATED_AT:
                return SessionOrders.created_at(ascending=ascending)
            case ComputeSessionOrderField.ID:
                return SessionOrders.id(ascending=ascending)

        raise ValueError(f"Unknown order field: {order.field}")



from __future__ import annotations

import asyncio
import functools
import grp
import logging
import os
import pwd
import signal
import ssl
import sys
import traceback
from collections.abc import (
    AsyncGenerator,
    AsyncIterator,
    Callable,
    Iterable,
    Mapping,
    Sequence,
)
from contextlib import AsyncExitStack, asynccontextmanager
from dataclasses import dataclass
from datetime import UTC, datetime
from pathlib import Path
from typing import (
    Any,
    Final,
)

import aiomonitor
import aiotools
import click
import uvloop
from aiohttp import web
from aiohttp.typedefs import Handler, Middleware
from opentelemetry.instrumentation.aiohttp_server import (
    middleware as otel_server_middleware,
)
from setproctitle import setproctitle

from ai.backend.common.cli import LazyGroup
from ai.backend.common.config import find_config_file
from ai.backend.common.dependencies import DependencyBuilderStack
from ai.backend.common.json import dump_json_str
from ai.backend.common.metrics.http import build_prometheus_metrics_handler
from ai.backend.common.metrics.metric import CommonMetricRegistry
from ai.backend.common.metrics.multiprocess import cleanup_prometheus_multiprocess_dir
from ai.backend.common.msgpack import DEFAULT_PACK_OPTS, DEFAULT_UNPACK_OPTS
from ai.backend.common.utils import env_info
from ai.backend.logging import BraceStyleAdapter, Logger, LogLevel
from ai.backend.logging.otel import (
    instrument_aiohttp_client,
    instrument_aiohttp_server,
)

from . import __version__
from .api import ManagerStatus
from .api.rest.middleware import (
    build_auth_middleware,
    build_exception_middleware,
)
from .config.bootstrap import BootstrapConfig
from .config.unified import EventLoopType
from .dependencies import DependencyInput, DependencyResources, ManagerDependencyComposer
from .plugin.webapp import WebappPluginContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

EVENT_DISPATCHER_CONSUMER_GROUP: Final = "manager"


@asynccontextmanager
async def webapp_plugin_ctx(
    root_app: web.Application,
    *,
    dep_resources: DependencyResources,
    pidx: int,
) -> AsyncIterator[None]:
    r = dep_resources
    plugin_ctx = WebappPluginContext(
        r.bootstrap.etcd,
        r.bootstrap.config_provider.config.model_dump(by_alias=True),
    )
    await plugin_ctx.init(
        context=None,
        allowlist=r.bootstrap.config_provider.config.manager.allowed_plugins,
        blocklist=r.bootstrap.config_provider.config.manager.disabled_plugins,
    )
    cors_options = r.system.cors_options
    for plugin_name, plugin_instance in plugin_ctx.plugins.items():
        if pidx == 0:
            log.info("Loading webapp plugin: {0}", plugin_name)
        subapp, global_middlewares = await plugin_instance.create_app(cors_options)
        _init_subapp(plugin_name, root_app, subapp, global_middlewares)
    try:
        yield
    finally:
        await plugin_ctx.cleanup()


@asynccontextmanager
async def manager_status_ctx(
    pidx: int,
    config_provider: Any,
) -> AsyncIterator[None]:
    if pidx == 0:
        mgr_status = await config_provider.legacy_etcd_config_loader.get_manager_status()
        if mgr_status is None or mgr_status not in (ManagerStatus.RUNNING, ManagerStatus.FROZEN):
            # legacy transition: we now have only RUNNING or FROZEN for HA setup.
            await config_provider.legacy_etcd_config_loader.update_manager_status(
                ManagerStatus.RUNNING
            )
            mgr_status = ManagerStatus.RUNNING
        log.info("Manager status: {}", mgr_status)
        tz = config_provider.config.system.timezone
        log.info("Configured timezone: {}", tz.tzname(datetime.now(UTC)))
    yield


_error_monitor_ref: Any = None


def handle_loop_error(
    loop: asyncio.AbstractEventLoop,
    context: Mapping[str, Any],
) -> None:
    exception = context.get("exception")
    msg = context.get("message", "(empty message)")
    if exception is not None:
        if sys.exc_info()[0] is not None:
            log.exception("Error inside event loop: {0}", msg)
            if _error_monitor_ref is not None:
                loop.create_task(_error_monitor_ref.capture_exception())
        else:
            exc_info = (type(exception), exception, exception.__traceback__)
            log.error("Error inside event loop: {0}", msg, exc_info=exc_info)
            if _error_monitor_ref is not None:
                loop.create_task(_error_monitor_ref.capture_exception(exc_instance=exception))


def _init_subapp(
    pkg_name: str,
    root_app: web.Application,
    subapp: web.Application,
    global_middlewares: Iterable[Middleware],
) -> None:
    from .api.rest.app import on_prepare

    subapp.on_response_prepare.append(on_prepare)

    async def _set_root_app(subapp: web.Application) -> None:
        subapp["_root_app"] = root_app

    # We must copy the public interface prior to all user-defined startup signal handlers.
    subapp.on_startup.insert(0, _set_root_app)
    if "prefix" not in subapp:
        subapp["prefix"] = pkg_name.split(".")[-1].replace("_", "-")
    prefix = subapp["prefix"]
    root_app.add_subapp("/" + prefix, subapp)
    root_app.middlewares.extend(global_middlewares)


def build_prometheus_service_discovery_handler(
    service_discovery: Any,
) -> Handler:
    async def _handler(_request: web.Request) -> web.Response:
        services = await service_discovery.discover()
        resp = []
        for service in services:
            resp.append({
                "targets": [f"{service.endpoint.prometheus_address}"],
                "labels": {
                    **service.labels,
                    "service_id": service.id,
                    "service_group": service.service_group,
                    "display_name": service.display_name,
                    "version": service.version,
                },
            })

        return web.json_response(
            resp,
            status=200,
            dumps=dump_json_str,
        )

    return _handler


def build_internal_app(dep_resources: DependencyResources) -> web.Application:
    app = web.Application()
    metric_registry = CommonMetricRegistry.instance()
    app.router.add_route("GET", r"/metrics", build_prometheus_metrics_handler(metric_registry))
    app.router.add_route(
        "GET",
        r"/metrics/service_discovery",
        build_prometheus_service_discovery_handler(dep_resources.system.service_discovery),
    )
    return app


@dataclass
class ServerMainArgs:
    bootstrap_cfg: BootstrapConfig
    bootstrap_cfg_path: Path
    log_endpoint: str
    log_level: LogLevel


@asynccontextmanager
async def server_main(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    args: ServerMainArgs,
) -> AsyncIterator[None]:
    boostrap_config = args.bootstrap_cfg
    loop.set_debug(boostrap_config.debug.asyncio)
    manager_init_stack = AsyncExitStack()

    @asynccontextmanager
    async def aiomonitor_ctx() -> AsyncIterator[aiomonitor.Monitor]:
        # Port is set by config where the defaults are:
        # termui_port = 38100 + pidx
        # webui_port = 39100 + pidx
        m = aiomonitor.Monitor(
            loop,
            termui_port=boostrap_config.manager.aiomonitor_termui_port + pidx,
            webui_port=boostrap_config.manager.aiomonitor_webui_port + pidx,
            console_enabled=False,
            hook_task_factory=boostrap_config.debug.enhanced_aiomonitor_task_info,
        )
        m.prompt = f"monitor (manager[{pidx}@{os.getpid()}]) >>> "
        m.console_locals["root_app"] = root_app
        aiomon_started = False
        try:
            m.start()
            aiomon_started = True
        except Exception as e:
            log.warning(
                "aiomonitor could not start but skipping this error to continue",
                exc_info=e,
            )
        try:
            yield m
        finally:
            if aiomon_started:
                m.close()

    @asynccontextmanager
    async def webapp_ctx(
        root_app: web.Application,
        dep_resources: DependencyResources,
    ) -> AsyncGenerator[None]:
        config_provider = dep_resources.bootstrap.config_provider

        runner = web.AppRunner(root_app, keepalive_timeout=30.0)

        internal_app = build_internal_app(dep_resources)
        internal_runner = web.AppRunner(internal_app, keepalive_timeout=30.0)

        ssl_ctx = None
        if config_provider.config.manager.ssl_enabled:
            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(
                str(config_provider.config.manager.ssl_cert),
                config_provider.config.manager.ssl_privkey,
            )
        await runner.setup()  # The cleanup context initialization happens here.
        await internal_runner.setup()
        service_addr = config_provider.config.manager.service_addr
        internal_addr = config_provider.config.manager.internal_addr
        site = web.TCPSite(
            runner,
            service_addr.host,
            service_addr.port,
            backlog=1024,
            reuse_port=True,
            ssl_context=ssl_ctx,
        )
        internal_site = web.TCPSite(
            internal_runner,
            internal_addr.host,
            internal_addr.port,
            backlog=1024,
            reuse_port=True,
        )
        await site.start()
        await internal_site.start()
        log.info(
            "started handling API requests at {}",
            service_addr,
        )

        try:
            yield
        finally:
            await runner.cleanup()

    await manager_init_stack.__aenter__()
    try:
        from .api.rest.app import build_root_app

        root_app = build_root_app(
            pidx,
            boostrap_config,
            loop_error_handler=functools.partial(handle_loop_error),
        )

        await manager_init_stack.enter_async_context(aiomonitor_ctx())

        # Initialize all dependencies via the Composer (replaces individual cleanup contexts)
        dep_stack = DependencyBuilderStack()
        await manager_init_stack.enter_async_context(dep_stack)
        dep_input = DependencyInput(
            config_path=args.bootstrap_cfg_path,
            pidx=pidx,
            log_level=args.log_level,
        )
        dep_resources = await dep_stack.enter_composer(
            ManagerDependencyComposer(),
            dep_input,
        )

        # Set the error monitor reference for the event loop error handler
        global _error_monitor_ref
        _error_monitor_ref = dep_resources.monitoring.error_monitor

        # Insert DI-based middlewares now that dependencies are available.
        # Maintain order: request_id(0) → exception(1) → auth(2) → api → metric
        root_app.middlewares.insert(
            1,
            build_exception_middleware(
                error_monitor=dep_resources.monitoring.error_monitor,
                stats_monitor=dep_resources.monitoring.stats_monitor,
                config_provider=dep_resources.bootstrap.config_provider,
            ),
        )
        root_app.middlewares.insert(
            2,
            build_auth_middleware(
                db=dep_resources.infrastructure.db,
                jwt_validator=dep_resources.system.jwt_validator,
                valkey_stat=dep_resources.infrastructure.valkey.stat,
                hook_plugin_ctx=dep_resources.plugins.hook_plugin_ctx,
            ),
        )

        # Build and mount the API module tree.
        # Must happen before runner.setup() which freezes the application router.
        from .api.rest.setup import setup_api

        setup_api(root_app, dep_resources, pidx)

        # Manager status check
        config_provider = dep_resources.bootstrap.config_provider
        await manager_init_stack.enter_async_context(manager_status_ctx(pidx, config_provider))

        # TODO: Remove manual middleware injection once the manager startup is
        # decoupled from the aiohttp Application lifecycle. Currently root_app is
        # instantiated before OTel config is available, so instrument_aiohttp_server()
        # (which patches the class via setattr) cannot take effect automatically.
        if config_provider.config.otel.enabled:
            instrument_aiohttp_server()
            instrument_aiohttp_client()
            root_app.middlewares.insert(0, otel_server_middleware)

        # Plugin webapps should be loaded before runner.setup() because root_app is frozen upon on_startup event.
        await manager_init_stack.enter_async_context(
            webapp_plugin_ctx(root_app, dep_resources=dep_resources, pidx=pidx)
        )
        await manager_init_stack.enter_async_context(webapp_ctx(root_app, dep_resources))

        if os.geteuid() == 0:
            uid = config_provider.config.manager.user
            gid = config_provider.config.manager.group
            if uid is None or gid is None:
                raise ValueError("user/group must be specified when running as root")

            os.setgroups([
                g.gr_gid for g in grp.getgrall() if pwd.getpwuid(uid).pw_name in g.gr_mem
            ])
            os.setgid(gid)
            os.setuid(uid)
            log.info("changed process uid and gid to {}:{}", uid, gid)

        log.info("Started the manager service.")
    except Exception:
        log.exception("Server initialization failure; triggering shutdown...")
        loop.call_later(0.2, os.kill, 0, signal.SIGINT)

    try:
        yield
    finally:
        _error_monitor_ref = None
        log.info("shutting down...")
        await manager_init_stack.__aexit__(None, None, None)


@aiotools.server_context
async def server_main_logwrapper(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    tuple_args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    setproctitle(f"backend.ai: manager worker-{pidx}")
    args = ServerMainArgs(
        bootstrap_cfg=tuple_args[0],
        bootstrap_cfg_path=tuple_args[1],
        log_endpoint=tuple_args[2],
        log_level=tuple_args[3],
    )
    logger = Logger(
        args.bootstrap_cfg.logging,
        is_master=False,
        log_endpoint=args.log_endpoint,
        msgpack_options={
            "pack_opts": DEFAULT_PACK_OPTS,
            "unpack_opts": DEFAULT_UNPACK_OPTS,
        },
    )
    try:
        with logger:
            async with server_main(loop, pidx, args):
                yield
    except Exception:
        traceback.print_exc(file=sys.stderr)


@click.group(invoke_without_command=True)
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=Path,
    default=None,
    help="The config file path. (default: ./manager.toml and /etc/backend.ai/manager.toml)",
)
@click.option(
    "--debug",
    is_flag=True,
    help="A shortcut to set `--log-level=DEBUG`",
)
@click.option(
    "--log-level",
    type=click.Choice([*LogLevel], case_sensitive=False),
    default=LogLevel.NOTSET,
    help="Set the logging verbosity level",
)
@click.pass_context
def main(
    ctx: click.Context,
    config_path: Path | None,
    debug: bool,
    log_level: LogLevel,
) -> None:
    """
    Start the manager service as a foreground process.
    """
    log_level = LogLevel.DEBUG if debug else log_level

    if config_path is None:
        discovered_cfg_path = find_config_file("manager")
    else:
        discovered_cfg_path = Path(config_path)

    bootstrap_cfg = asyncio.run(BootstrapConfig.load_from_file(discovered_cfg_path, log_level))

    if ctx.invoked_subcommand is None:
        bootstrap_cfg.manager.pid_file.write_text(str(os.getpid()))
        ipc_base_path = bootstrap_cfg.manager.ipc_base_path
        log_sockpath = ipc_base_path / f"manager-logger-{os.getpid()}.sock"
        log_endpoint = f"ipc://{log_sockpath}"
        try:
            logger = Logger(
                bootstrap_cfg.logging,
                is_master=True,
                log_endpoint=log_endpoint,
                msgpack_options={
                    "pack_opts": DEFAULT_PACK_OPTS,
                    "unpack_opts": DEFAULT_UNPACK_OPTS,
                },
            )
            with logger:
                ns = bootstrap_cfg.etcd.namespace
                setproctitle(f"backend.ai: manager {ns}")
                log.info("Backend.AI Manager {0}", __version__)
                log.info("runtime: {0}", env_info())
                log_config = logging.getLogger("ai.backend.manager.config")
                log_config.debug("debug mode enabled.")
                runner: Callable[..., Any]
                match bootstrap_cfg.manager.event_loop:
                    case EventLoopType.UVLOOP:
                        runner = uvloop.run
                        log.info("Using uvloop as the event loop backend")
                    case EventLoopType.ASYNCIO:
                        runner = asyncio.run
                try:
                    aiotools.start_server(
                        server_main_logwrapper,
                        num_workers=bootstrap_cfg.manager.num_proc,
                        args=(bootstrap_cfg, discovered_cfg_path, log_endpoint, log_level),
                        wait_timeout=5.0,
                        runner=runner,
                    )
                finally:
                    cleanup_prometheus_multiprocess_dir()
                    log.info("terminated.")
        finally:
            if bootstrap_cfg.manager.pid_file.is_file():
                # check is_file() to prevent deleting /dev/null!
                bootstrap_cfg.manager.pid_file.unlink()
    else:
        # Click is going to invoke a subcommand.
        pass


@main.group(cls=LazyGroup, import_name="ai.backend.manager.api.auth:cli")
def auth() -> None:
    pass


if __name__ == "__main__":
    sys.exit(main())



"""
Manager-specific path parameter models for user admin REST API.
"""

import uuid

from pydantic import Field

from ai.backend.common.api_handlers import BaseRequestModel


class GetUserPathParam(BaseRequestModel):
    user_id: uuid.UUID = Field(description="The user ID to retrieve")


class UpdateUserPathParam(BaseRequestModel):
    user_id: uuid.UUID = Field(description="The user ID to update")



"""Prometheus query preset GQL types."""

from .filters import (
    QueryDefinitionFilter,
    QueryDefinitionOrderBy,
    QueryDefinitionOrderField,
)
from .inputs import (
    CreateQueryDefinitionInput,
    ExecuteQueryDefinitionOptionsInput,
    MetricLabelEntryInput,
    ModifyQueryDefinitionInput,
    QueryTimeRangeInput,
)
from .node import (
    CreateQueryDefinitionPayload,
    ModifyQueryDefinitionPayload,
    QueryDefinitionConnection,
    QueryDefinitionEdge,
    QueryDefinitionGQL,
)
from .payloads import (
    DeleteQueryDefinitionPayload,
    MetricLabelEntryGQL,
    MetricResultGQL,
    MetricResultValueGQL,
    QueryDefinitionOptionsGQL,
    QueryDefinitionResultGQL,
)

__all__ = [
    # Node types
    "QueryDefinitionGQL",
    "QueryDefinitionEdge",
    "QueryDefinitionConnection",
    # Filter and OrderBy
    "QueryDefinitionFilter",
    "QueryDefinitionOrderBy",
    "QueryDefinitionOrderField",
    # Input types
    "CreateQueryDefinitionInput",
    "ModifyQueryDefinitionInput",
    "QueryTimeRangeInput",
    "MetricLabelEntryInput",
    "ExecuteQueryDefinitionOptionsInput",
    # Payload and result types
    "QueryDefinitionOptionsGQL",
    "MetricLabelEntryGQL",
    "MetricResultValueGQL",
    "MetricResultGQL",
    "QueryDefinitionResultGQL",
    "CreateQueryDefinitionPayload",
    "ModifyQueryDefinitionPayload",
    "DeleteQueryDefinitionPayload",
]



from __future__ import annotations

import uuid
from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType

from .base import ErrorLogAction


@dataclass
class MarkClearedErrorLogAction(ErrorLogAction):
    """Action to mark an error log as cleared."""

    log_id: uuid.UUID
    user_uuid: uuid.UUID
    user_domain: str
    is_superadmin: bool
    is_admin: bool

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE

    @override
    def entity_id(self) -> str | None:
        return str(self.log_id)


@dataclass
class MarkClearedErrorLogActionResult(BaseActionResult):
    """Result of marking an error log as cleared."""

    @override
    def entity_id(self) -> str | None:
        return None



from __future__ import annotations

import logging
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

from ai.backend.common.plugin.hook import ALL_COMPLETED, PASSED, HookPluginContext
from ai.backend.logging import BraceStyleAdapter

from .base import PluginDependency, PluginsInput

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class HookPluginDependency(PluginDependency[HookPluginContext]):
    """Provides HookPluginContext lifecycle management.

    After initialization, dispatches ACTIVATE_MANAGER and raises
    RuntimeError if the hook result is not PASSED.
    """

    @property
    def stage_name(self) -> str:
        return "hook-plugin"

    @asynccontextmanager
    async def provide(self, setup_input: PluginsInput) -> AsyncIterator[HookPluginContext]:
        """Initialize and provide a HookPluginContext.

        Args:
            setup_input: Plugins input containing etcd and config

        Yields:
            Initialized HookPluginContext

        Raises:
            RuntimeError: If ACTIVATE_MANAGER hook dispatch does not pass
        """
        ctx = HookPluginContext(setup_input.etcd, setup_input.local_config)
        await ctx.init(
            context=setup_input.init_context,
            allowlist=setup_input.allowed_plugins,
            blocklist=setup_input.disabled_plugins,
        )
        hook_result = await ctx.dispatch(
            "ACTIVATE_MANAGER",
            (),
            return_when=ALL_COMPLETED,
        )
        if hook_result.status != PASSED:
            raise RuntimeError("Could not activate the manager instance.")
        log.info("HookPluginContext initialized with plugins: {}", list(ctx.plugins.keys()))
        try:
            yield ctx
        finally:
            await ctx.cleanup()



from __future__ import annotations

import uuid
from dataclasses import dataclass, field
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.models.user import UserRole

from .base import TemplateAction
from .create_task_template import TaskTemplateItemInput


@dataclass
class UpdateTaskTemplateAction(TemplateAction):
    """Action to update an existing task template."""

    template_id: str
    domain_name: str
    requesting_group: str
    requester_uuid: uuid.UUID
    requester_access_key: str
    requester_role: UserRole
    requester_domain: str
    owner_access_key: str | None
    items: list[TaskTemplateItemInput] = field(default_factory=list)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE

    @override
    def entity_id(self) -> str | None:
        return self.template_id


@dataclass
class UpdateTaskTemplateActionResult(BaseActionResult):
    """Result of updating a task template."""

    @override
    def entity_id(self) -> str | None:
        return None



"""Artifact registry handler class using constructor dependency injection.

All handlers use the new ApiHandler pattern: typed parameters
(``BodyParam``, ``PathParam``, ``QueryParam``) are automatically
extracted by ``_wrap_api_handler`` and responses are returned as
``APIResponse`` objects.
"""

from __future__ import annotations

import logging
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, Final, cast

import sqlalchemy as sa

from ai.backend.common.api_handlers import APIResponse, BodyParam, PathParam, QueryParam
from ai.backend.common.data.storage.registries.types import ModelSortKey, ModelTarget
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.artifact.types import (
    ArtifactDataWithRevisionsResponse,
    ArtifactFilterOptions,
    ArtifactOrderingOptions,
    ArtifactRevisionResponseData,
)
from ai.backend.manager.dto.request import (
    DelegateImportArtifactsReq,
    DelegateScanArtifactsReq,
    ScanArtifactModelPathParam,
    ScanArtifactModelQueryParam,
    ScanArtifactModelsReq,
    ScanArtifactsReq,
    SearchArtifactsReq,
)
from ai.backend.manager.dto.response import (
    ArtifactRevisionImportTask,
    DelegateImportArtifactsResponse,
    DelegateScanArtifactsResponse,
    RetreiveArtifactModelResponse,
    ScanArtifactModelsResponse,
    ScanArtifactsResponse,
    SearchArtifactsResponse,
)
from ai.backend.manager.errors.artifact import ArtifactImportDelegationError
from ai.backend.manager.models.artifact import ArtifactRow
from ai.backend.manager.repositories.base import (
    BatchQuerier,
    OffsetPagination,
    QueryCondition,
    QueryOrder,
)
from ai.backend.manager.services.artifact.actions.delegate_scan import DelegateScanArtifactsAction
from ai.backend.manager.services.artifact.actions.retrieve_model import RetrieveModelAction
from ai.backend.manager.services.artifact.actions.retrieve_model_multi import (
    RetrieveModelsAction,
)
from ai.backend.manager.services.artifact.actions.scan import ScanArtifactsAction
from ai.backend.manager.services.artifact.actions.search_with_revisions import (
    SearchArtifactsWithRevisionsAction,
)
from ai.backend.manager.services.artifact_revision.actions.delegate_import_revision_batch import (
    DelegateImportArtifactRevisionBatchAction,
)

if TYPE_CHECKING:
    from ai.backend.manager.services.artifact.processors import ArtifactProcessors
    from ai.backend.manager.services.artifact_revision.processors import ArtifactRevisionProcessors

log: Final = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ArtifactRegistryHandler:
    """Artifact registry API handler with constructor-injected dependencies."""

    def __init__(
        self,
        *,
        artifact: ArtifactProcessors,
        artifact_revision: ArtifactRevisionProcessors,
    ) -> None:
        self._artifact = artifact
        self._artifact_revision = artifact_revision

    async def scan_artifacts(
        self,
        body: BodyParam[ScanArtifactsReq],
    ) -> APIResponse:
        """
        Scan external registries to discover available artifacts.

        Searches HuggingFace or Reservoir registries for artifacts matching the specified
        criteria and registers them in the system with SCANNED status. The artifacts
        become available for import but are not downloaded until explicitly imported.

        This is the first step in the artifact workflow: Scan -> Import -> Use.
        """

        action_result = await self._artifact.scan.wait_for_complete(
            ScanArtifactsAction(
                registry_id=body.parsed.registry_id,
                artifact_type=body.parsed.artifact_type,
                limit=body.parsed.limit,
                order=ModelSortKey.DOWNLOADS,
                search=body.parsed.search,
            )
        )

        resp = ScanArtifactsResponse(
            artifacts=[
                ArtifactDataWithRevisionsResponse.from_artifact_with_revisions(artifact)
                for artifact in action_result.result
            ],
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    async def delegate_scan_artifacts(
        self,
        body: BodyParam[DelegateScanArtifactsReq],
    ) -> APIResponse:
        action_result = await self._artifact.delegate_scan.wait_for_complete(
            DelegateScanArtifactsAction(
                delegator_reservoir_id=body.parsed.delegator_reservoir_id,
                artifact_type=body.parsed.artifact_type,
                search=body.parsed.search,
                order=ModelSortKey.DOWNLOADS,
                delegatee_target=body.parsed.delegatee_target
                if body.parsed.delegatee_target
                else None,
                limit=body.parsed.limit,
            )
        )

        resp = DelegateScanArtifactsResponse(
            artifacts=[
                ArtifactDataWithRevisionsResponse.from_artifact_with_revisions(artifact)
                for artifact in action_result.result
            ],
            source_registry_id=action_result.source_registry_id,
            source_registry_type=action_result.source_registry_type,
            readme_data=action_result.readme_data,
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    async def delegate_import_artifacts(
        self,
        body: BodyParam[DelegateImportArtifactsReq],
    ) -> APIResponse:
        force = body.parsed.options.force
        action_result = (
            await self._artifact_revision.delegate_import_revision_batch.wait_for_complete(
                DelegateImportArtifactRevisionBatchAction(
                    delegator_reservoir_id=body.parsed.delegator_reservoir_id,
                    artifact_type=body.parsed.artifact_type,
                    delegatee_target=body.parsed.delegatee_target
                    if body.parsed.delegatee_target
                    else None,
                    artifact_revision_ids=body.parsed.artifact_revision_ids,
                    force=force,
                )
            )
        )

        if len(action_result.result) != len(action_result.task_ids):
            raise ArtifactImportDelegationError(
                "Mismatch between artifact revisions and task IDs returned"
            )

        # Convert to ArtifactRevisionImportTask format
        tasks = []
        for task_uuid, revision in zip(action_result.task_ids, action_result.result, strict=True):
            task_id = str(task_uuid) if task_uuid is not None else None
            tasks.append(
                ArtifactRevisionImportTask(
                    task_id=task_id,
                    artifact_revision=ArtifactRevisionResponseData.from_revision_data(revision),
                )
            )

        resp = DelegateImportArtifactsResponse(tasks=tasks)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    async def search_artifacts(
        self,
        body: BodyParam[SearchArtifactsReq],
    ) -> APIResponse:
        """
        Search registered artifacts with offset-based pagination.

        Returns artifacts that have been previously scanned and registered in the system.
        Supports efficient pagination for browsing through large datasets of artifacts
        with their revision information.
        """

        pagination_opts = body.parsed.pagination
        filters = body.parsed.filters
        ordering = body.parsed.ordering

        # Build BatchQuerier from REST pagination options
        offset_opts = pagination_opts.offset
        pagination = OffsetPagination(
            limit=offset_opts.limit if offset_opts and offset_opts.limit is not None else 20,
            offset=offset_opts.offset if offset_opts and offset_opts.offset is not None else 0,
        )

        # Convert filter options to query conditions
        conditions: list[QueryCondition] = []
        if filters is not None:
            conditions.extend(_build_artifact_filter_conditions(filters))

        # Convert ordering options to query orders
        orders: list[QueryOrder] = []
        if ordering is not None:
            orders.extend(_build_artifact_query_orders(ordering))

        querier = BatchQuerier(
            pagination=pagination,
            conditions=conditions,
            orders=orders,
        )

        action_result = await self._artifact.search_artifacts_with_revisions.wait_for_complete(
            SearchArtifactsWithRevisionsAction(querier=querier)
        )

        artifacts = action_result.data
        resp = SearchArtifactsResponse(
            artifacts=[
                ArtifactDataWithRevisionsResponse.from_artifact_with_revisions(artifact)
                for artifact in artifacts
            ],
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    async def scan_models(
        self,
        body: BodyParam[ScanArtifactModelsReq],
    ) -> APIResponse:
        """
        Perform batch scanning of specific models from external registries.

        Scans multiple specified models and retrieves detailed information.
        The README content and file sizes are processed in the background,
        unlike single model scanning which retrieves this information immediately.
        """

        action_result = await self._artifact.retrieve_models.wait_for_complete(
            RetrieveModelsAction(
                models=body.parsed.models,
                registry_id=body.parsed.registry_id,
            )
        )

        resp = ScanArtifactModelsResponse(
            artifacts=[
                ArtifactDataWithRevisionsResponse.from_artifact_with_revisions(artifact)
                for artifact in action_result.result
            ],
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    async def scan_single_model(
        self,
        path: PathParam[ScanArtifactModelPathParam],
        query: QueryParam[ScanArtifactModelQueryParam],
    ) -> APIResponse:
        """
        Scan a single model and retrieve detailed information immediately.

        Performs immediate detailed scanning of a specified model including
        README content and file sizes. This provides complete metadata
        for the model, ready for import or direct use.
        """

        model = ModelTarget(
            model_id=path.parsed.model_id,
            revision=query.parsed.revision,
        )
        action_result = await self._artifact.retrieve_single_model.wait_for_complete(
            RetrieveModelAction(
                model=model,
                registry_id=query.parsed.registry_id,
            )
        )

        resp = RetreiveArtifactModelResponse(artifact=action_result.result)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)


def _build_artifact_filter_conditions(
    filters: ArtifactFilterOptions,
) -> list[QueryCondition]:
    """Convert ArtifactFilterOptions to a list of QueryConditions for BatchQuerier."""
    conditions: list[QueryCondition] = []

    if filters.artifact_type:
        artifact_types = filters.artifact_type
        conditions.append(lambda: ArtifactRow.type.in_(artifact_types))

    if filters.name_filter is not None:
        name_filter = filters.name_filter

        def _name_cond() -> sa.sql.expression.ColumnElement[bool]:
            cond = name_filter.apply_to_column(
                cast(sa.sql.elements.ColumnElement[str], ArtifactRow.name)
            )
            if cond is None:
                return sa.literal(True)
            return cond

        conditions.append(_name_cond)

    if filters.registry_id is not None:
        registry_id = filters.registry_id
        conditions.append(lambda: ArtifactRow.registry_id == registry_id)

    if filters.registry_type is not None:
        registry_type = filters.registry_type
        conditions.append(lambda: ArtifactRow.registry_type == registry_type)

    if filters.source_registry_id is not None:
        source_registry_id = filters.source_registry_id
        conditions.append(lambda: ArtifactRow.source_registry_id == source_registry_id)

    if filters.source_registry_type is not None:
        source_registry_type = filters.source_registry_type
        conditions.append(lambda: ArtifactRow.source_registry_type == source_registry_type)

    if filters.availability:
        availability = filters.availability
        conditions.append(lambda: ArtifactRow.availability.in_(availability))

    return conditions


def _build_artifact_query_orders(
    ordering: ArtifactOrderingOptions,
) -> list[QueryOrder]:
    """Convert ArtifactOrderingOptions to a list of QueryOrders for BatchQuerier."""
    orders: list[QueryOrder] = []
    for field, desc in ordering.order_by:
        column: Any = getattr(ArtifactRow, field.value.lower(), ArtifactRow.name)
        if desc:
            orders.append(column.desc())
        else:
            orders.append(column.asc())
    return orders



"""Spec handler class using constructor dependency injection.

Serves OpenAPI and GraphiQL documentation endpoints.
These handlers return HTML or JSON responses directly (not APIResponse)
since they serve static documentation content.
"""

from __future__ import annotations

import logging
from http import HTTPStatus
from typing import TYPE_CHECKING, Final

from aiohttp import web

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.dto.context import RequestCtx
from ai.backend.manager.errors.common import GenericForbidden
from ai.backend.manager.openapi import generate_openapi

if TYPE_CHECKING:
    from aiohttp.web import Application

    from ai.backend.manager.config.provider import ManagerConfigProvider

log: Final = BraceStyleAdapter(logging.getLogger(__spec__.name))


OPENAPI_HTML = """
<!DOCTYPE html>
<html>
  <head>
    <title>Backend.AI REST API Reference</title>
    <meta charset="utf-8"/>
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <style>
      body {
        margin: 0;
        padding: 0;
      }
    </style>
  </head>
  <body>
    <redoc spec-url="openapi/spec.json"></redoc>
    <script src="../static/vendor/spec-viewer.js"></script>
  </body>
</html>
"""

GRAPHIQL_HTML = """
<html>
  <head>
    <title>Backend.AI GraphQL API Reference</title>
\t<meta charset="UTF-8">
    <link href="../static/vendor/graphiql.min.css" rel="stylesheet" />
  </head>
  <body style="margin: 0;">
    <div id="graphiql" style="height: 100vh;"></div>

    <script src="../static/vendor/react.production.min.js"
    ></script>
    <script src="../static/vendor/react-dom.production.min.js"
    ></script>
    <script src="../static/vendor/graphiql.min.js"
    ></script>

    <script>
      const fetcher = GraphiQL.createFetcher({ url: '../admin/gql' });

      ReactDOM.render(
        React.createElement(GraphiQL, { fetcher: fetcher }),
        document.getElementById('graphiql'),
      );
    </script>
  </body>
</html>
"""

GRAPHIQL_V2_HTML = """
<html>
  <head>
    <title>Backend.AI GraphQL V2 API Reference</title>
\t<meta charset="UTF-8">
    <link href="../../static/vendor/graphiql.min.css" rel="stylesheet" />
  </head>
  <body style="margin: 0;">
    <div id="graphiql" style="height: 100vh;"></div>

    <script src="../../static/vendor/react.production.min.js"
    ></script>
    <script src="../../static/vendor/react-dom.production.min.js"
    ></script>
    <script src="../../static/vendor/graphiql.min.js"
    ></script>

    <script>
      const fetcher = GraphiQL.createFetcher({
        url: '/func/admin/gql/strawberry',
        subscriptionUrl: '/func/admin/gql/strawberry'
      });

      ReactDOM.render(
        React.createElement(GraphiQL, { fetcher: fetcher }),
        document.getElementById('graphiql'),
      );
    </script>
  </body>
</html>
"""


class SpecHandler:
    """Spec API handler for documentation endpoints."""

    def __init__(
        self,
        *,
        config_provider: ManagerConfigProvider,
        root_app: Application,
    ) -> None:
        self._config_provider = config_provider
        self._root_app = root_app

    # ------------------------------------------------------------------
    # render_graphiql_graphene_html (GET /spec/graphiql)
    # ------------------------------------------------------------------

    async def render_graphiql_graphene_html(
        self,
        ctx: RequestCtx,
    ) -> web.StreamResponse:
        if not self._config_provider.config.api.allow_graphql_schema_introspection:
            raise GenericForbidden
        return web.Response(
            body=GRAPHIQL_HTML,
            status=HTTPStatus.OK,
            content_type="text/html",
        )

    # ------------------------------------------------------------------
    # render_graphiql_strawberry_html (GET /spec/graphiql/strawberry)
    # ------------------------------------------------------------------

    async def render_graphiql_strawberry_html(
        self,
        ctx: RequestCtx,
    ) -> web.StreamResponse:
        if not self._config_provider.config.api.allow_graphql_schema_introspection:
            raise GenericForbidden
        return web.Response(
            body=GRAPHIQL_V2_HTML,
            status=HTTPStatus.OK,
            content_type="text/html",
        )

    # ------------------------------------------------------------------
    # render_openapi_html (GET /spec/openapi)
    # ------------------------------------------------------------------

    async def render_openapi_html(
        self,
        ctx: RequestCtx,
    ) -> web.StreamResponse:
        if not self._config_provider.config.api.allow_openapi_schema_introspection:
            raise GenericForbidden
        return web.Response(
            body=OPENAPI_HTML,
            status=HTTPStatus.OK,
            content_type="text/html",
        )

    # ------------------------------------------------------------------
    # generate_openapi_spec (GET /spec/openapi/spec.json)
    # ------------------------------------------------------------------

    async def generate_openapi_spec(
        self,
        ctx: RequestCtx,
    ) -> web.StreamResponse:
        if not self._config_provider.config.api.allow_openapi_schema_introspection:
            raise GenericForbidden
        return web.json_response(generate_openapi(self._root_app._subapps))



"""Service (model serving) sub-app lifecycle hooks.

Extracted from the legacy ``api/service.py`` module so that the
``rest/service`` package owns its own startup/shutdown concerns.
"""

from __future__ import annotations

import logging

import aiotools
import attrs
from aiohttp import web

from ai.backend.logging import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@attrs.define(slots=True, auto_attribs=True, init=False)
class PrivateContext:
    database_ptask_group: aiotools.PersistentTaskGroup


async def init(app: web.Application) -> None:
    app_ctx: PrivateContext = app["services.context"]
    app_ctx.database_ptask_group = aiotools.PersistentTaskGroup()


async def shutdown(app: web.Application) -> None:
    app_ctx: PrivateContext = app["services.context"]
    await app_ctx.database_ptask_group.shutdown()



"""Resource module registrar."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.manager.api.rest.middleware.auth import auth_required, superadmin_required
from ai.backend.manager.api.rest.routing import RouteRegistry

from .handler import ResourceHandler

if TYPE_CHECKING:
    from ai.backend.manager.api.rest.types import RouteDeps


def register_resource_routes(handler: ResourceHandler, route_deps: RouteDeps) -> RouteRegistry:
    """Build the resource sub-application."""

    reg = RouteRegistry.create("resource", route_deps.cors_options)

    # Public preset listing (auth required)
    reg.add("GET", "/presets", handler.list_presets, middlewares=[auth_required])

    # Container registries (superadmin)
    reg.add(
        "GET",
        "/container-registries",
        handler.get_container_registries,
        middlewares=[superadmin_required],
    )

    # Check presets (auth required)
    reg.add("POST", "/check-presets", handler.check_presets, middlewares=[auth_required])

    # Recalculate usage (superadmin)
    reg.add(
        "POST",
        "/recalculate-usage",
        handler.recalculate_usage,
        middlewares=[superadmin_required],
    )

    # Usage statistics (superadmin)
    reg.add("GET", "/usage/month", handler.usage_per_month, middlewares=[superadmin_required])
    reg.add(
        "GET",
        "/usage/period",
        handler.usage_per_period,
        middlewares=[superadmin_required],
    )

    # User stats (auth required)
    reg.add("GET", "/stats/user/month", handler.user_month_stats, middlewares=[auth_required])

    # Admin stats (superadmin)
    reg.add(
        "GET",
        "/stats/admin/month",
        handler.admin_month_stats,
        middlewares=[superadmin_required],
    )

    # Watcher endpoints (superadmin)
    reg.add("GET", "/watcher", handler.get_watcher_status, middlewares=[superadmin_required])
    reg.add(
        "POST",
        "/watcher/agent/start",
        handler.watcher_agent_start,
        middlewares=[superadmin_required],
    )
    reg.add(
        "POST",
        "/watcher/agent/stop",
        handler.watcher_agent_stop,
        middlewares=[superadmin_required],
    )
    reg.add(
        "POST",
        "/watcher/agent/restart",
        handler.watcher_agent_restart,
        middlewares=[superadmin_required],
    )
    return reg



"""Handler for terminating routes."""

import logging
from collections.abc import Sequence

from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.deployment.types import RouteStatus, RouteStatusTransitions
from ai.backend.manager.defs import LockID
from ai.backend.manager.repositories.deployment.types import RouteData
from ai.backend.manager.sokovan.deployment.route.executor import RouteExecutor
from ai.backend.manager.sokovan.deployment.route.types import RouteExecutionResult

from .base import RouteHandler

log = BraceStyleAdapter(logging.getLogger(__name__))


class TerminatingRouteHandler(RouteHandler):
    """Handler for terminating routes (TERMINATING -> TERMINATED)."""

    def __init__(
        self,
        route_executor: RouteExecutor,
        event_producer: EventProducer,
    ) -> None:
        self._route_executor = route_executor
        self._event_producer = event_producer

    @classmethod
    def name(cls) -> str:
        """Get the name of the handler."""
        return "terminate-routes"

    @property
    def lock_id(self) -> LockID | None:
        """No lock needed for terminating routes."""
        return None

    @classmethod
    def target_statuses(cls) -> list[RouteStatus]:
        """Get the target route statuses for this handler."""
        return [RouteStatus.TERMINATING]

    @classmethod
    def next_status(cls) -> RouteStatus | None:
        """Get the next route status after this handler's operation."""
        return RouteStatus.TERMINATED

    @classmethod
    def failure_status(cls) -> RouteStatus | None:
        """Get the failure route status if applicable."""
        # Even if termination fails, we still mark as terminated
        return None

    @classmethod
    def stale_status(cls) -> RouteStatus | None:
        """Get the stale route status if applicable."""
        return None

    @classmethod
    def status_transitions(cls) -> RouteStatusTransitions:
        """Define state transitions for terminating route handler (BEP-1030).

        - success: Route → TERMINATED
        - failure: None (even if termination fails, route proceeds to terminated)
        - stale: None
        """
        return RouteStatusTransitions(
            success=RouteStatus.TERMINATED,
            failure=None,
            stale=None,
        )

    async def execute(self, routes: Sequence[RouteData]) -> RouteExecutionResult:
        """Execute termination for routes."""
        log.debug("Terminating {} routes", len(routes))

        # Execute route termination logic via executor
        return await self._route_executor.terminate_routes(routes)

    async def post_process(self, result: RouteExecutionResult) -> None:
        """Handle post-processing after terminating routes."""
        log.info(
            "Terminated {} routes successfully, {} failed",
            len(result.successes),
            len(result.errors),
        )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, override

from ai.backend.common.types import AccessKey
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.services.stream.actions.base import StreamAction


@dataclass(frozen=True)
class ExecuteInStreamAction(StreamAction):
    session_name: str
    access_key: AccessKey
    api_version: tuple[int, str]
    run_id: str
    mode: str
    code: str
    opts: dict[str, Any] = field(default_factory=dict)
    flush_timeout: float | None = None

    @override
    def entity_id(self) -> str | None:
        return self.session_name

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass(frozen=True)
class ExecuteInStreamActionResult(BaseActionResult):
    result: dict[str, Any]

    @override
    def entity_id(self) -> str | None:
        return None



from __future__ import annotations

import logging
import socket
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass

from ai.backend.common.clients.valkey_client.valkey_leader.client import ValkeyLeaderClient
from ai.backend.common.defs import REDIS_STREAM_LOCK, RedisRole
from ai.backend.common.dependencies import NonMonitorableDependencyProvider
from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.common.events.event_types.artifact_registry.anycast import (
    DoScanReservoirRegistryEvent,
)
from ai.backend.common.leader import ValkeyLeaderElection, ValkeyLeaderElectionConfig
from ai.backend.common.leader.tasks import EventProducerTask, LeaderCron, PeriodicTask
from ai.backend.common.leader.tasks.event_task import EventTaskSpec
from ai.backend.common.types import ValkeyProfileTarget
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.sokovan.sokovan import SokovanOrchestrator

log = BraceStyleAdapter(logging.getLogger(__name__))


@dataclass
class LeaderElectionInput:
    """Input required for leader election setup."""

    valkey_profile_target: ValkeyProfileTarget
    pidx: int
    config_provider: ManagerConfigProvider
    event_producer: EventProducer
    sokovan_orchestrator: SokovanOrchestrator


class LeaderElectionDependency(
    NonMonitorableDependencyProvider[LeaderElectionInput, ValkeyLeaderElection]
):
    """Provides ValkeyLeaderElection lifecycle management.

    Wraps the leader election initialization from the original
    ``leader_election_ctx`` in server.py. Creates leader client,
    configures election, collects task specs, and manages the
    election lifecycle.
    """

    @property
    def stage_name(self) -> str:
        return "leader-election"

    @asynccontextmanager
    async def provide(
        self, setup_input: LeaderElectionInput
    ) -> AsyncIterator[ValkeyLeaderElection]:
        """Initialize and provide leader election.

        Args:
            setup_input: Input containing valkey target, config, event producer,
                and sokovan orchestrator for task spec collection

        Yields:
            Started ValkeyLeaderElection instance
        """
        # Create ValkeyLeaderClient for leader election
        valkey_leader_client = await ValkeyLeaderClient.create(
            valkey_target=setup_input.valkey_profile_target.profile_target(RedisRole.STREAM),
            db_id=REDIS_STREAM_LOCK,
            human_readable_name="leader",
        )

        # Create leader election configuration
        server_id = f"manager-{socket.gethostname()}-{setup_input.pidx}"
        leader_config = ValkeyLeaderElectionConfig(
            server_id=server_id,
            leader_key="leader:sokovan:scheduler",
            lease_duration=30,
            renewal_interval=10.0,
            failure_threshold=3,
        )

        # Create leader election instance
        leader_election = ValkeyLeaderElection(
            leader_client=valkey_leader_client,
            config=leader_config,
        )

        # Get task specifications from sokovan and register them
        task_specs = setup_input.sokovan_orchestrator.create_task_specs()

        # Rescan reservoir registry periodically
        reservoir_config = setup_input.config_provider.config.reservoir

        if reservoir_config and reservoir_config.use_delegation:
            task_specs.append(
                EventTaskSpec(
                    name="reservoir_registry_scan",
                    event_factory=lambda: DoScanReservoirRegistryEvent(),
                    interval=600,
                    initial_delay=0,
                )
            )

        # Create event producer tasks from specs
        leader_tasks: list[PeriodicTask] = [
            EventProducerTask(spec, setup_input.event_producer) for spec in task_specs
        ]

        # Register tasks with the election system
        leader_cron = LeaderCron(tasks=leader_tasks)
        leader_election.register_task(leader_cron)

        # Start leader election
        await leader_election.start()
        log.info("Leader election started for server {}", server_id)

        try:
            yield leader_election
        finally:
            await leader_election.stop()



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.user.types import UserData
from ai.backend.manager.repositories.base.querier import BatchQuerier
from ai.backend.manager.repositories.user.types import RoleUserSearchScope
from ai.backend.manager.services.user.actions.base import UserAction


@dataclass
class SearchUsersByRoleAction(UserAction):
    """Action for searching users assigned to a role."""

    scope: RoleUserSearchScope
    querier: BatchQuerier

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH


@dataclass
class SearchUsersByRoleActionResult(BaseActionResult):
    """Result of searching users assigned to a role."""

    users: list[UserData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool

    @override
    def entity_id(self) -> str | None:
        return None



from __future__ import annotations

from ai.backend.manager.data.prometheus_query_preset import (
    PrometheusQueryPresetData,
    PrometheusQueryPresetListResult,
    PrometheusQueryPresetModifier,
)

from .creators import PrometheusQueryPresetCreatorSpec
from .repositories import PrometheusQueryPresetRepositories
from .repository import PrometheusQueryPresetRepository

__all__ = (
    "PrometheusQueryPresetCreatorSpec",
    "PrometheusQueryPresetData",
    "PrometheusQueryPresetListResult",
    "PrometheusQueryPresetModifier",
    "PrometheusQueryPresetRepositories",
    "PrometheusQueryPresetRepository",
)



from __future__ import annotations

import logging
from collections.abc import Callable
from typing import Any, Final

from graphql import GraphQLResolveInfo
from strawberry.extensions.base_extension import SchemaExtension
from strawberry.utils.await_maybe import AwaitableOrValue

from ai.backend.logging.utils import BraceStyleAdapter

log: Final = BraceStyleAdapter(logging.getLogger(__spec__.name))


class GQLLoggingExtension(SchemaExtension):
    """Logs GraphQL operation details for audit purposes."""

    def resolve(
        self,
        _next: Callable[..., Any],
        root: Any,
        info: GraphQLResolveInfo,
        *args: Any,
        **kwargs: Any,
    ) -> AwaitableOrValue[object]:
        if info.path.prev is None:
            # TODO: Log access_key when StrawberryGQLContext has user info
            log.info(
                "ADMIN.GQL.V2 ({}:{}, op:{})",
                info.operation.operation,
                info.field_name,
                info.operation.name,
            )
        result: AwaitableOrValue[object] = _next(root, info, *args, **kwargs)
        return result



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType

from .base import ManagerAdminAction


@dataclass
class UpdateManagerStatusAction(ManagerAdminAction):
    """Action to update the manager status."""

    status: str
    force_kill: bool

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class UpdateManagerStatusActionResult(BaseActionResult):
    """Result of updating manager status."""

    @override
    def entity_id(self) -> str | None:
        return None



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.scaling_group.repository import ScalingGroupRepository
from ai.backend.manager.repositories.types import RepositoryArgs


@dataclass
class ScalingGroupRepositories:
    repository: ScalingGroupRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = ScalingGroupRepository(
            args.db,
        )

        return cls(
            repository=repository,
        )



"""Fair share module registrar."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.manager.api.rest.middleware.auth import auth_required, superadmin_required
from ai.backend.manager.api.rest.routing import RouteRegistry

from .handler import FairShareAPIHandler

if TYPE_CHECKING:
    from ai.backend.manager.api.rest.types import RouteDeps


def register_fair_share_routes(
    handler: FairShareAPIHandler, route_deps: RouteDeps
) -> RouteRegistry:
    """Build the fair share sub-application."""

    reg = RouteRegistry.create("fair-share", route_deps.cors_options)

    # Domain fair share routes (superadmin)
    reg.add(
        "GET",
        "/domains/{resource_group}/{domain_name}",
        handler.get_domain_fair_share,
        middlewares=[superadmin_required],
    )
    reg.add(
        "POST",
        "/domains/search",
        handler.search_domain_fair_shares,
        middlewares=[superadmin_required],
    )

    # Project fair share routes (superadmin)
    reg.add(
        "GET",
        "/projects/{resource_group}/{project_id}",
        handler.get_project_fair_share,
        middlewares=[superadmin_required],
    )
    reg.add(
        "POST",
        "/projects/search",
        handler.search_project_fair_shares,
        middlewares=[superadmin_required],
    )

    # User fair share routes (superadmin)
    reg.add(
        "GET",
        "/users/{resource_group}/{project_id}/{user_uuid}",
        handler.get_user_fair_share,
        middlewares=[superadmin_required],
    )
    reg.add(
        "POST",
        "/users/search",
        handler.search_user_fair_shares,
        middlewares=[superadmin_required],
    )

    # Usage bucket routes (superadmin)
    reg.add(
        "POST",
        "/usage-buckets/domains/search",
        handler.search_domain_usage_buckets,
        middlewares=[superadmin_required],
    )
    reg.add(
        "POST",
        "/usage-buckets/projects/search",
        handler.search_project_usage_buckets,
        middlewares=[superadmin_required],
    )
    reg.add(
        "POST",
        "/usage-buckets/users/search",
        handler.search_user_usage_buckets,
        middlewares=[superadmin_required],
    )

    # RG-scoped usage bucket routes (auth required)
    reg.add(
        "POST",
        "/rg/{resource_group}/usage-buckets/domains/search",
        handler.rg_search_domain_usage_buckets,
        middlewares=[auth_required],
    )
    reg.add(
        "POST",
        "/rg/{resource_group}/domains/{domain_name}/usage-buckets/projects/search",
        handler.rg_search_project_usage_buckets,
        middlewares=[auth_required],
    )
    reg.add(
        "POST",
        "/rg/{resource_group}/domains/{domain_name}/projects/{project_id}/usage-buckets/users/search",
        handler.rg_search_user_usage_buckets,
        middlewares=[auth_required],
    )

    # RG-scoped domain fair share routes (auth required)
    reg.add(
        "GET",
        "/rg/{resource_group}/domains/{domain_name}",
        handler.rg_get_domain_fair_share,
        middlewares=[auth_required],
    )
    reg.add(
        "POST",
        "/rg/{resource_group}/domains/search",
        handler.rg_search_domain_fair_shares,
        middlewares=[auth_required],
    )

    # RG-scoped project fair share routes (auth required)
    reg.add(
        "GET",
        "/rg/{resource_group}/domains/{domain_name}/projects/{project_id}",
        handler.rg_get_project_fair_share,
        middlewares=[auth_required],
    )
    reg.add(
        "POST",
        "/rg/{resource_group}/domains/{domain_name}/projects/search",
        handler.rg_search_project_fair_shares,
        middlewares=[auth_required],
    )

    # RG-scoped user fair share routes (auth required)
    reg.add(
        "GET",
        "/rg/{resource_group}/domains/{domain_name}/projects/{project_id}/users/{user_uuid}",
        handler.rg_get_user_fair_share,
        middlewares=[auth_required],
    )
    reg.add(
        "POST",
        "/rg/{resource_group}/domains/{domain_name}/projects/{project_id}/users/search",
        handler.rg_search_user_fair_shares,
        middlewares=[auth_required],
    )

    # Upsert weight routes (superadmin)
    reg.add(
        "PUT",
        "/domains/{resource_group}/{domain_name}/weight",
        handler.upsert_domain_fair_share_weight,
        middlewares=[superadmin_required],
    )
    reg.add(
        "PUT",
        "/projects/{resource_group}/{project_id}/weight",
        handler.upsert_project_fair_share_weight,
        middlewares=[superadmin_required],
    )
    reg.add(
        "PUT",
        "/users/{resource_group}/{project_id}/{user_uuid}/weight",
        handler.upsert_user_fair_share_weight,
        middlewares=[superadmin_required],
    )

    # Bulk upsert weight routes (superadmin)
    reg.add(
        "POST",
        "/domains/bulk-upsert-weight",
        handler.bulk_upsert_domain_fair_share_weight,
        middlewares=[superadmin_required],
    )
    reg.add(
        "POST",
        "/projects/bulk-upsert-weight",
        handler.bulk_upsert_project_fair_share_weight,
        middlewares=[superadmin_required],
    )
    reg.add(
        "POST",
        "/users/bulk-upsert-weight",
        handler.bulk_upsert_user_fair_share_weight,
        middlewares=[superadmin_required],
    )

    # Resource group spec routes (superadmin)
    reg.add(
        "GET",
        "/resource-groups/{resource_group}/spec",
        handler.get_resource_group_fair_share_spec,
        middlewares=[superadmin_required],
    )
    reg.add(
        "GET",
        "/resource-groups/specs",
        handler.search_resource_group_fair_share_specs,
        middlewares=[superadmin_required],
    )
    reg.add(
        "PATCH",
        "/resource-groups/{resource_group}/spec",
        handler.update_resource_group_fair_share_spec,
        middlewares=[superadmin_required],
    )
    return reg



"""RBAC sub-registry registrar."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.manager.api.rest.middleware.auth import auth_required
from ai.backend.manager.api.rest.routing import RouteRegistry

from .handler import RBACHandler

if TYPE_CHECKING:
    from ai.backend.manager.api.rest.types import RouteDeps


def register_rbac_routes(handler: RBACHandler, route_deps: RouteDeps) -> RouteRegistry:
    """Build the RBAC sub-registry (child of admin)."""
    reg = RouteRegistry.create("rbac", route_deps.cors_options)

    # Role management routes
    reg.add("POST", "/roles", handler.create_role, middlewares=[auth_required])
    reg.add("POST", "/roles/search", handler.search_roles, middlewares=[auth_required])
    reg.add("GET", "/roles/{role_id}", handler.get_role, middlewares=[auth_required])
    reg.add("PATCH", "/roles/{role_id}", handler.update_role, middlewares=[auth_required])
    reg.add("POST", "/roles/delete", handler.delete_role, middlewares=[auth_required])
    reg.add("POST", "/roles/purge", handler.purge_role, middlewares=[auth_required])

    # Role assignment routes
    reg.add("POST", "/roles/assign", handler.assign_role, middlewares=[auth_required])
    reg.add("POST", "/roles/revoke", handler.revoke_role, middlewares=[auth_required])
    reg.add(
        "POST",
        "/roles/{role_id}/assigned-users/search",
        handler.search_assigned_users,
        middlewares=[auth_required],
    )

    # Scope routes
    reg.add("GET", "/scope-types", handler.get_scope_types, middlewares=[auth_required])
    reg.add(
        "POST",
        "/scopes/{scope_type}/search",
        handler.search_scopes,
        middlewares=[auth_required],
    )

    # Entity routes
    reg.add("GET", "/entity-types", handler.get_entity_types, middlewares=[auth_required])
    reg.add(
        "POST",
        "/scopes/{scope_type}/{scope_id}/entities/{entity_type}/search",
        handler.search_entities,
        middlewares=[auth_required],
    )

    return reg



from __future__ import annotations

from .actions.search import SearchServiceCatalogsAction, SearchServiceCatalogsActionResult
from .processors import ServiceCatalogProcessors
from .service import ServiceCatalogService

__all__ = (
    "SearchServiceCatalogsAction",
    "SearchServiceCatalogsActionResult",
    "ServiceCatalogProcessors",
    "ServiceCatalogService",
)



import uuid
from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.services.auth.actions.base import AuthAction


@dataclass
class ResolveUserScopeAction(AuthAction):
    requester_uuid: uuid.UUID
    requester_role: UserRole
    requester_domain: str
    is_superadmin: bool
    owner_user_email: str | None  # None = self

    @override
    def entity_id(self) -> str | None:
        return str(self.requester_uuid)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET


@dataclass
class ResolveUserScopeResult(BaseActionResult):
    owner_uuid: uuid.UUID
    owner_role: UserRole

    @override
    def entity_id(self) -> str | None:
        return None



"""
Adapter for converting service search DTOs to repository query objects.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.manager.api.rest.adapter import BaseFilterAdapter
from ai.backend.manager.repositories.base import QueryCondition
from ai.backend.manager.repositories.deployment.options import DeploymentConditions

if TYPE_CHECKING:
    from ai.backend.common.dto.manager.model_serving.request import ServiceFilterModel


class ServiceSearchAdapter(BaseFilterAdapter):
    """Adapter for converting service search filters to query conditions."""

    def convert_filter(self, filter: ServiceFilterModel) -> list[QueryCondition]:
        """Convert ServiceFilterModel to a list of query conditions."""
        conditions: list[QueryCondition] = []

        if filter.name is not None:
            condition = self.convert_string_filter(
                filter.name,
                contains_factory=DeploymentConditions.by_name_contains,
                equals_factory=DeploymentConditions.by_name_equals,
                starts_with_factory=DeploymentConditions.by_name_starts_with,
                ends_with_factory=DeploymentConditions.by_name_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        return conditions



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType

from .base import DotfileAction


@dataclass
class UpdateBootstrapScriptAction(DotfileAction):
    """Action to update a user's bootstrap script."""

    access_key: str
    script: str

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE

    @override
    def entity_id(self) -> str | None:
        return self.access_key


@dataclass
class UpdateBootstrapScriptActionResult(BaseActionResult):
    """Result of updating a bootstrap script."""

    @override
    def entity_id(self) -> str | None:
        return None



import uuid
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager as actxmgr
from datetime import UTC, datetime
from typing import Any, cast

import sqlalchemy as sa
from sqlalchemy.engine import CursorResult
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.ext.asyncio import async_sessionmaker
from sqlalchemy.orm import selectinload

from ai.backend.common.data.artifact.types import ArtifactRegistryType, VerificationStepResult
from ai.backend.common.data.permission.types import RBACElementType
from ai.backend.common.data.storage.registries.types import ModelData
from ai.backend.common.data.storage.types import ArtifactStorageType
from ai.backend.manager.data.artifact.types import (
    ArtifactAvailability,
    ArtifactData,
    ArtifactDataWithRevisions,
    ArtifactListResult,
    ArtifactRemoteStatus,
    ArtifactRevisionData,
    ArtifactRevisionListResult,
    ArtifactStatus,
    ArtifactType,
    ArtifactWithRevisionsListResult,
)
from ai.backend.manager.data.association.types import AssociationArtifactsStoragesData
from ai.backend.manager.data.permission.types import RBACElementRef
from ai.backend.manager.errors.artifact import (
    ArtifactAssociationDeletionError,
    ArtifactAssociationNotFoundError,
    ArtifactNotFoundError,
    ArtifactNotVerified,
    ArtifactRevisionNotFoundError,
    ArtifactUpdateError,
)
from ai.backend.manager.models.artifact import ArtifactRow
from ai.backend.manager.models.artifact_revision import ArtifactRevisionRow
from ai.backend.manager.models.association_artifacts_storages import AssociationArtifactsStorageRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.repositories.artifact.creators import ArtifactCreatorSpec
from ai.backend.manager.repositories.base import BatchQuerier, execute_batch_querier
from ai.backend.manager.repositories.base.rbac.entity_creator import (
    RBACEntityCreator,
    execute_rbac_entity_creator,
)
from ai.backend.manager.repositories.base.updater import Updater, execute_updater


class ArtifactDBSource:
    """Database source for artifact-related operations."""

    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    async def get_artifact_by_id(self, artifact_id: uuid.UUID) -> ArtifactData:
        async with self._db.begin_readonly_session_read_committed() as db_sess:
            result = await db_sess.execute(
                sa.select(ArtifactRow).where(ArtifactRow.id == artifact_id)
            )
            row = result.scalar_one_or_none()
            if row is None:
                raise ArtifactNotFoundError(f"Artifact with ID {artifact_id} not found")
            return row.to_dataclass()

    async def get_artifact_revision_by_id(self, revision_id: uuid.UUID) -> ArtifactRevisionData:
        async with self._db.begin_readonly_session_read_committed() as db_sess:
            result = await db_sess.execute(
                sa.select(ArtifactRevisionRow).where(ArtifactRevisionRow.id == revision_id)
            )
            row = result.scalar_one_or_none()
            if row is None:
                raise ArtifactRevisionNotFoundError(
                    f"Artifact revision with ID {revision_id} not found"
                )
            return row.to_dataclass()

    async def get_model_artifact(self, model_id: str, registry_id: uuid.UUID) -> ArtifactData:
        async with self._db.begin_readonly_session_read_committed() as db_sess:
            result = await db_sess.execute(
                sa.select(ArtifactRow).where(
                    sa.and_(ArtifactRow.name == model_id, ArtifactRow.registry_id == registry_id)
                )
            )
            row = result.scalar_one_or_none()
            if row is None:
                raise ArtifactNotFoundError(
                    f"Artifact with model ID {model_id} not found under registry {registry_id}"
                )
            return row.to_dataclass()

    async def get_artifact_revision(
        self, artifact_id: uuid.UUID, revision: str
    ) -> ArtifactRevisionData:
        async with self._db.begin_readonly_session_read_committed() as db_sess:
            result = await db_sess.execute(
                sa.select(ArtifactRevisionRow).where(
                    sa.and_(
                        ArtifactRevisionRow.artifact_id == artifact_id,
                        ArtifactRevisionRow.version == revision,
                    )
                )
            )
            row = result.scalar_one_or_none()
            if row is None:
                raise ArtifactRevisionNotFoundError(f"Revision {revision} not found")
            return row.to_dataclass()

    async def update_artifact(self, updater: Updater[ArtifactRow]) -> ArtifactData:
        async with self._db.begin_session() as db_sess:
            # Check if artifact exists and is not deleted
            check_result = await db_sess.execute(
                sa.select(ArtifactRow.id).where(
                    sa.and_(
                        ArtifactRow.id == updater.pk_value,
                        ArtifactRow.availability != ArtifactAvailability.DELETED,
                    )
                )
            )
            if check_result.scalar_one_or_none() is None:
                raise ArtifactNotFoundError(f"Artifact with ID {updater.pk_value} not found")

            result = await execute_updater(db_sess, updater)
            if result is None:
                raise ArtifactNotFoundError(f"Artifact with ID {updater.pk_value} not found")
            return result.row.to_dataclass()

    async def list_artifact_revisions(self, artifact_id: uuid.UUID) -> list[ArtifactRevisionData]:
        async with self._db.begin_readonly_session_read_committed() as db_sess:
            result = await db_sess.execute(
                sa.select(ArtifactRevisionRow).where(ArtifactRevisionRow.artifact_id == artifact_id)
            )
            rows = list(result.scalars().all())
            return [row.to_dataclass() for row in rows]

    # TODO: Refactor using on_conflict_do_update?
    async def upsert_artifacts(
        self,
        artifacts: list[ArtifactData],
    ) -> list[ArtifactData]:
        async with self._db.begin_session() as db_sess:
            result_artifacts: list[ArtifactData] = []

            for artifact_data in artifacts:
                # Check if artifact exists
                artifact_query_result = await db_sess.execute(
                    sa.select(ArtifactRow).where(
                        sa.and_(
                            ArtifactRow.name == artifact_data.name,
                            ArtifactRow.registry_id == artifact_data.registry_id,
                        )
                    )
                )
                existing_artifact = artifact_query_result.scalar_one_or_none()

                if existing_artifact is None:
                    # Create new artifact with RBAC scope association
                    rbac_creator = RBACEntityCreator(
                        spec=ArtifactCreatorSpec(
                            name=artifact_data.name,
                            type=artifact_data.type,
                            description=artifact_data.description,
                            registry_id=artifact_data.registry_id,
                            registry_type=artifact_data.registry_type,
                            source_registry_id=artifact_data.source_registry_id,
                            source_registry_type=artifact_data.source_registry_type,
                            readonly=True,
                            extra=artifact_data.extra,
                        ),
                        element_type=RBACElementType.ARTIFACT,
                        scope_ref=RBACElementRef(
                            RBACElementType.ARTIFACT_REGISTRY,
                            str(artifact_data.registry_id),
                        ),
                    )
                    creator_result = await execute_rbac_entity_creator(db_sess, rbac_creator)
                    new_artifact = creator_result.row
                    await db_sess.refresh(
                        new_artifact, attribute_names=["scanned_at", "updated_at"]
                    )
                    result_artifacts.append(new_artifact.to_dataclass())
                else:
                    # Update existing artifact
                    has_changes = (existing_artifact.description != artifact_data.description) or (
                        existing_artifact.extra != artifact_data.extra
                    )
                    if has_changes:
                        existing_artifact.extra = artifact_data.extra
                        existing_artifact.description = artifact_data.description
                        existing_artifact.updated_at = datetime.now(UTC)

                    await db_sess.flush()
                    await db_sess.refresh(
                        existing_artifact, attribute_names=["scanned_at", "updated_at"]
                    )
                    result_artifacts.append(existing_artifact.to_dataclass())

            return result_artifacts

    async def upsert_artifact_revisions(
        self,
        revisions: list[ArtifactRevisionData],
    ) -> list[ArtifactRevisionData]:
        async with self._db.begin_session() as db_sess:
            result_revisions: list[ArtifactRevisionData] = []
            artifact_ids_to_update: set[uuid.UUID] = set()

            for revision_data in revisions:
                # Skip failed or rejected revision copy
                if revision_data.status in [ArtifactStatus.FAILED, ArtifactStatus.REJECTED]:
                    continue

                # Check if revision exists
                revision_query_result = await db_sess.execute(
                    sa.select(ArtifactRevisionRow).where(
                        sa.and_(
                            ArtifactRevisionRow.artifact_id == revision_data.artifact_id,
                            ArtifactRevisionRow.version == revision_data.version,
                        )
                    )
                )
                existing_revision = revision_query_result.scalar_one_or_none()

                if existing_revision is None:
                    # Create new revision
                    verification_result = None
                    if revision_data.verification_result is not None:
                        verification_result = revision_data.verification_result.model_dump()

                    new_revision = ArtifactRevisionRow(
                        id=revision_data.id,
                        artifact_id=revision_data.artifact_id,
                        version=revision_data.version,
                        readme=revision_data.readme,
                        size=revision_data.size,
                        status=ArtifactStatus.SCANNED,
                        remote_status=revision_data.remote_status,
                        created_at=revision_data.created_at,
                        updated_at=revision_data.updated_at,
                        digest=revision_data.digest,
                        verification_result=verification_result,
                    )
                    db_sess.add(new_revision)
                    await db_sess.flush()
                    await db_sess.refresh(new_revision)
                    result_revisions.append(new_revision.to_dataclass())
                    artifact_ids_to_update.add(revision_data.artifact_id)
                else:
                    # Update existing revision only if there are changes
                    verification_result = None
                    if revision_data.verification_result is not None:
                        verification_result = revision_data.verification_result.model_dump()

                    has_changes = (
                        existing_revision.readme != revision_data.readme
                        or existing_revision.digest != revision_data.digest
                        or existing_revision.remote_status != revision_data.remote_status
                        or existing_revision.verification_result != verification_result,
                    )

                    if has_changes:
                        existing_revision.readme = revision_data.readme
                        existing_revision.size = revision_data.size
                        existing_revision.created_at = revision_data.created_at
                        existing_revision.updated_at = revision_data.updated_at
                        existing_revision.digest = revision_data.digest
                        existing_revision.verification_result = verification_result

                        # This must be done to avoid overwriting local revisions' remote_status with None
                        if revision_data.remote_status is not None:
                            existing_revision.remote_status = revision_data.remote_status
                        artifact_ids_to_update.add(revision_data.artifact_id)

                    await db_sess.flush()
                    await db_sess.refresh(existing_revision)
                    result_revisions.append(existing_revision.to_dataclass())

            # Update artifact updated_at timestamp for affected artifacts
            if artifact_ids_to_update:
                await db_sess.execute(
                    sa.update(ArtifactRow)
                    .where(ArtifactRow.id.in_(artifact_ids_to_update))
                    .values(updated_at=sa.func.now())
                )

            return result_revisions

    async def upsert_huggingface_model_artifacts(
        self,
        model_list: list[ModelData],
        registry_id: uuid.UUID,
    ) -> list[ArtifactDataWithRevisions]:
        async with self._db.begin_session() as db_sess:
            # key: artifact_id
            artifacts_map: dict[uuid.UUID, tuple[ArtifactRow, list[ArtifactRevisionRow]]] = {}
            artifact_ids_to_update: set[uuid.UUID] = set()

            for model in model_list:
                # Check if artifact exists within the current session
                artifact_query_result = await db_sess.execute(
                    sa.select(ArtifactRow).where(
                        sa.and_(
                            ArtifactRow.name == model.id, ArtifactRow.registry_id == registry_id
                        )
                    )
                )
                artifact_row = artifact_query_result.scalar_one_or_none()

                if artifact_row is None:
                    # Create new artifact with RBAC scope association
                    rbac_creator = RBACEntityCreator(
                        spec=ArtifactCreatorSpec(
                            name=model.id,
                            type=ArtifactType.MODEL,
                            registry_id=registry_id,
                            registry_type=ArtifactRegistryType.HUGGINGFACE,
                            source_registry_id=registry_id,
                            source_registry_type=ArtifactRegistryType.HUGGINGFACE,
                            readonly=True,
                            extra=model.extra,
                        ),
                        element_type=RBACElementType.ARTIFACT,
                        scope_ref=RBACElementRef(
                            RBACElementType.ARTIFACT_REGISTRY,
                            str(registry_id),
                        ),
                    )
                    creator_result = await execute_rbac_entity_creator(db_sess, rbac_creator)
                    artifact_row = creator_result.row
                    await db_sess.refresh(
                        artifact_row, attribute_names=["scanned_at", "updated_at"]
                    )
                else:
                    # Update existing artifact's extra field if changed
                    artifact_row.extra = model.extra
                    artifact_ids_to_update.add(artifact_row.id)

                # Initialize artifact in map if not exists
                if artifact_row.id not in artifacts_map:
                    artifacts_map[artifact_row.id] = (artifact_row, [])

                # Check if artifact revision exists
                revision_query_result = await db_sess.execute(
                    sa.select(ArtifactRevisionRow).where(
                        sa.and_(
                            ArtifactRevisionRow.artifact_id == artifact_row.id,
                            ArtifactRevisionRow.version == model.revision,
                        )
                    )
                )

                existing_revision = revision_query_result.scalar_one_or_none()
                if existing_revision is not None:
                    # Update existing revision only if there are changes
                    has_changes = existing_revision.digest != model.sha

                    if has_changes:
                        existing_revision.readme = model.readme
                        existing_revision.updated_at = model.modified_at
                        artifact_ids_to_update.add(artifact_row.id)

                    await db_sess.flush()
                    await db_sess.refresh(existing_revision)
                    artifacts_map[artifact_row.id][1].append(existing_revision)
                else:
                    # Insert new artifact revision
                    new_revision = ArtifactRevisionRow.from_huggingface_model_data(
                        artifact_id=artifact_row.id,
                        model_data=model,
                    )

                    db_sess.add(new_revision)
                    await db_sess.flush()
                    await db_sess.refresh(new_revision)
                    artifacts_map[artifact_row.id][1].append(new_revision)
                    artifact_ids_to_update.add(artifact_row.id)

            # Update artifact updated_at timestamp for affected artifacts
            if artifact_ids_to_update:
                await db_sess.execute(
                    sa.update(ArtifactRow)
                    .where(ArtifactRow.id.in_(artifact_ids_to_update))
                    .values(updated_at=sa.func.now())
                )

                # Refresh updated_at for affected artifacts
                for artifact_id, (artifact_row, _revs) in artifacts_map.items():
                    if artifact_id in artifact_ids_to_update:
                        await db_sess.refresh(artifact_row, attribute_names=["updated_at"])

            # Convert to ArtifactDataWithRevisions format
            result: list[ArtifactDataWithRevisions] = []
            for artifact_row, revision_rows in artifacts_map.values():
                artifact_data = artifact_row.to_dataclass()
                revision_data_list = [revision.to_dataclass() for revision in revision_rows]
                result.append(
                    ArtifactDataWithRevisions.from_dataclasses(
                        artifact_data=artifact_data, revisions=revision_data_list
                    )
                )

        return result

    async def associate_artifact_with_storage(
        self,
        artifact_revision_id: uuid.UUID,
        storage_namespace_id: uuid.UUID,
        storage_type: ArtifactStorageType,
    ) -> AssociationArtifactsStoragesData:
        async with self._db.begin_session() as db_sess:
            select_stmt = sa.select(AssociationArtifactsStorageRow.id).where(
                sa.and_(
                    AssociationArtifactsStorageRow.artifact_revision_id == artifact_revision_id,
                    AssociationArtifactsStorageRow.storage_namespace_id == storage_namespace_id,
                )
            )
            existing = (await db_sess.execute(select_stmt)).scalar_one_or_none()
            if existing is not None:
                return AssociationArtifactsStoragesData(
                    id=existing,
                    artifact_revision_id=artifact_revision_id,
                    storage_namespace_id=storage_namespace_id,
                )

            insert_stmt = (
                sa.insert(AssociationArtifactsStorageRow)
                .values(
                    artifact_revision_id=artifact_revision_id,
                    storage_namespace_id=storage_namespace_id,
                    storage_type=storage_type.value,
                )
                .returning(AssociationArtifactsStorageRow.id)
            )

            result = await db_sess.execute(insert_stmt)
            existing = result.scalar_one()

            return AssociationArtifactsStoragesData(
                id=existing,
                artifact_revision_id=artifact_revision_id,
                storage_namespace_id=storage_namespace_id,
            )

    async def disassociate_artifact_with_storage(
        self, artifact_revision_id: uuid.UUID, storage_namespace_id: uuid.UUID
    ) -> AssociationArtifactsStoragesData:
        async with self._db.begin_session() as db_sess:
            select_result = await db_sess.execute(
                sa.select(AssociationArtifactsStorageRow).where(
                    sa.and_(
                        AssociationArtifactsStorageRow.artifact_revision_id == artifact_revision_id,
                        AssociationArtifactsStorageRow.storage_namespace_id == storage_namespace_id,
                    )
                )
            )
            existing_row = select_result.scalar_one_or_none()
            if existing_row is None:
                raise ArtifactAssociationNotFoundError(
                    f"Association between artifact {artifact_revision_id} and storage {storage_namespace_id} does not exist"
                )

            # Store the data before deletion
            association_data = AssociationArtifactsStoragesData(
                id=existing_row.id,
                artifact_revision_id=existing_row.artifact_revision_id,
                storage_namespace_id=existing_row.storage_namespace_id,
            )

            # Delete the association
            delete_result = await db_sess.execute(
                sa.delete(AssociationArtifactsStorageRow).where(
                    sa.and_(
                        AssociationArtifactsStorageRow.artifact_revision_id == artifact_revision_id,
                        AssociationArtifactsStorageRow.storage_namespace_id == storage_namespace_id,
                    )
                )
            )

            if cast(CursorResult[Any], delete_result).rowcount == 0:
                raise ArtifactAssociationDeletionError("Failed to delete association")

            return association_data

    async def approve_artifact(self, revision_id: uuid.UUID) -> ArtifactRevisionData:
        async with self._db.begin_session() as db_sess:
            result = await db_sess.execute(
                sa.select(ArtifactRevisionRow).where(ArtifactRevisionRow.id == revision_id)
            )
            row = result.scalar_one_or_none()
            if row is None:
                raise ArtifactRevisionNotFoundError()

            if row.status == ArtifactStatus.AVAILABLE:
                raise ArtifactNotVerified("Artifacts already approved")
            if row.status != ArtifactStatus.NEEDS_APPROVAL:
                raise ArtifactNotVerified("Only verified artifacts could be approved")

            update_stmt = (
                sa.update(ArtifactRevisionRow)
                .where(
                    sa.and_(
                        ArtifactRevisionRow.id == revision_id,
                        ArtifactRevisionRow.status == ArtifactStatus.NEEDS_APPROVAL,
                    )
                )
                .values(status=ArtifactStatus.AVAILABLE)
                .returning(ArtifactRevisionRow)
            )

            result = await db_sess.execute(update_stmt)
            updated_row = result.scalars().one_or_none()
            if updated_row is None:
                raise ArtifactUpdateError()

            return updated_row.to_dataclass()

    async def reject_artifact(self, revision_id: uuid.UUID) -> ArtifactRevisionData:
        async with self._db.begin_session() as db_sess:
            result = await db_sess.execute(
                sa.select(ArtifactRevisionRow).where(ArtifactRevisionRow.id == revision_id)
            )
            row = result.scalar_one_or_none()
            if row is None:
                raise ArtifactRevisionNotFoundError()

            update_stmt = (
                sa.update(ArtifactRevisionRow)
                .where(ArtifactRevisionRow.id == revision_id)
                .values(status=ArtifactStatus.REJECTED.value)
                .returning(ArtifactRevisionRow)
            )

            result = await db_sess.execute(update_stmt)
            updated_row = result.scalars().one_or_none()
            if updated_row is None:
                raise ArtifactUpdateError()

            return updated_row.to_dataclass()

    async def reset_artifact_revision_status(self, revision_id: uuid.UUID) -> uuid.UUID:
        async with self._db.begin_session() as db_sess:
            stmt = (
                sa.update(ArtifactRevisionRow)
                .where(ArtifactRevisionRow.id == revision_id)
                .values(status=ArtifactStatus.SCANNED)
            )
            await db_sess.execute(stmt)
            return revision_id

    async def update_artifact_revision_status(
        self, artifact_revision_id: uuid.UUID, status: ArtifactStatus
    ) -> uuid.UUID:
        async with self._db.begin_session() as db_sess:
            stmt = (
                sa.update(ArtifactRevisionRow)
                .where(ArtifactRevisionRow.id == artifact_revision_id)
                .values(status=status)
            )
            await db_sess.execute(stmt)
            return artifact_revision_id

    async def update_artifact_revision_remote_status(
        self, artifact_revision_id: uuid.UUID, remote_status: ArtifactRemoteStatus
    ) -> uuid.UUID:
        async with self._db.begin_session() as db_sess:
            stmt = (
                sa.update(ArtifactRevisionRow)
                .where(ArtifactRevisionRow.id == artifact_revision_id)
                .values(remote_status=remote_status)
            )
            await db_sess.execute(stmt)
            return artifact_revision_id

    async def delete_artifacts(self, artifact_ids: list[uuid.UUID]) -> list[ArtifactData]:
        async with self._db.begin_session() as db_sess:
            # Update availability to DELETED for the given artifact IDs (only for ALIVE artifacts)
            await db_sess.execute(
                sa.update(ArtifactRow)
                .where(
                    sa.and_(
                        ArtifactRow.id.in_(artifact_ids),
                        ArtifactRow.availability != ArtifactAvailability.DELETED,
                    )
                )
                .values(availability=ArtifactAvailability.DELETED.value)
            )

            # Fetch and return the updated artifacts
            result = await db_sess.execute(
                sa.select(ArtifactRow).where(ArtifactRow.id.in_(artifact_ids))
            )
            rows = list(result.scalars().all())
            return [row.to_dataclass() for row in rows]

    async def restore_artifacts(self, artifact_ids: list[uuid.UUID]) -> list[ArtifactData]:
        async with self._db.begin_session() as db_sess:
            # Update availability to ALIVE for the given artifact IDs (only for DELETED artifacts)
            await db_sess.execute(
                sa.update(ArtifactRow)
                .where(
                    sa.and_(
                        ArtifactRow.id.in_(artifact_ids),
                        ArtifactRow.availability == ArtifactAvailability.DELETED,
                    )
                )
                .values(availability=ArtifactAvailability.ALIVE.value)
            )

            # Fetch and return the updated artifacts
            result = await db_sess.execute(
                sa.select(ArtifactRow).where(ArtifactRow.id.in_(artifact_ids))
            )
            rows = list(result.scalars().all())
            return [row.to_dataclass() for row in rows]

    async def update_artifact_revision_bytesize(
        self, artifact_revision_id: uuid.UUID, size: int
    ) -> uuid.UUID:
        async with self._begin_session_read_committed() as db_sess:
            stmt = (
                sa.update(ArtifactRevisionRow)
                .where(ArtifactRevisionRow.id == artifact_revision_id)
                .values(size=size)
            )
            await db_sess.execute(stmt)
            return artifact_revision_id

    async def update_artifact_revision_readme(
        self, artifact_revision_id: uuid.UUID, readme: str
    ) -> uuid.UUID:
        async with self._begin_session_read_committed() as db_sess:
            stmt = (
                sa.update(ArtifactRevisionRow)
                .where(ArtifactRevisionRow.id == artifact_revision_id)
                .values(readme=readme)
            )
            await db_sess.execute(stmt)
            return artifact_revision_id

    async def update_artifact_revision_verification_result(
        self,
        artifact_revision_id: uuid.UUID,
        verification_result: VerificationStepResult,
    ) -> uuid.UUID:
        async with self._begin_session_read_committed() as db_sess:
            stmt = (
                sa.update(ArtifactRevisionRow)
                .where(ArtifactRevisionRow.id == artifact_revision_id)
                .values(verification_result=verification_result.model_dump())
            )
            await db_sess.execute(stmt)
            return artifact_revision_id

    async def update_artifact_revision_digest(
        self, artifact_revision_id: uuid.UUID, digest: str
    ) -> uuid.UUID:
        async with self._begin_session_read_committed() as db_sess:
            stmt = (
                sa.update(ArtifactRevisionRow)
                .where(ArtifactRevisionRow.id == artifact_revision_id)
                .values(digest=digest)
            )
            await db_sess.execute(stmt)
            return artifact_revision_id

    async def get_artifact_revision_readme(self, artifact_revision_id: uuid.UUID) -> str | None:
        async with self._db.begin_readonly_session_read_committed() as db_sess:
            result = await db_sess.execute(
                sa.select(ArtifactRevisionRow.readme).where(
                    ArtifactRevisionRow.id == artifact_revision_id
                )
            )
            return result.scalar_one_or_none()

    async def search_artifacts(
        self,
        querier: BatchQuerier,
    ) -> ArtifactListResult:
        """Search artifacts with querier pattern.

        Args:
            querier: BatchQuerier for filtering, ordering, and pagination

        Returns:
            ArtifactListResult with items, total count, and pagination info
        """
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(ArtifactRow)

            result = await execute_batch_querier(
                db_sess,
                query,
                querier,
            )

            items = [row.ArtifactRow.to_dataclass() for row in result.rows]

            return ArtifactListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )

    async def search_artifact_revisions(
        self,
        querier: BatchQuerier,
    ) -> ArtifactRevisionListResult:
        """Search artifact revisions with querier pattern.

        Args:
            querier: BatchQuerier for filtering, ordering, and pagination

        Returns:
            ArtifactRevisionListResult with items, total count, and pagination info
        """
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(ArtifactRevisionRow)

            result = await execute_batch_querier(
                db_sess,
                query,
                querier,
            )

            items = [row.ArtifactRevisionRow.to_dataclass() for row in result.rows]

            return ArtifactRevisionListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )

    async def search_artifacts_with_revisions(
        self,
        querier: BatchQuerier,
    ) -> ArtifactWithRevisionsListResult:
        """Search artifacts with their revisions using querier pattern.

        Args:
            querier: BatchQuerier for filtering, ordering, and pagination

        Returns:
            ArtifactWithRevisionsListResult with items, total count, and pagination info
        """
        async with self._db.begin_readonly_session() as db_sess:
            # Build query with eager loading of revisions
            query = sa.select(ArtifactRow).options(selectinload(ArtifactRow.revision_rows))

            result = await execute_batch_querier(
                db_sess,
                query,
                querier,
            )

            # Convert to ArtifactDataWithRevisions objects
            items: list[ArtifactDataWithRevisions] = []
            for row in result.rows:
                artifact_data = row.ArtifactRow.to_dataclass()
                revisions_data = [
                    revision.to_dataclass() for revision in row.ArtifactRow.revision_rows
                ]
                items.append(
                    ArtifactDataWithRevisions.from_dataclasses(
                        artifact_data=artifact_data, revisions=revisions_data
                    )
                )

            return ArtifactWithRevisionsListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )

    @actxmgr
    async def _begin_session_read_committed(self) -> AsyncIterator[SASession]:
        """
        Begin a read-write session with READ COMMITTED isolation level.
        """
        async with self._db.connect() as conn:
            # Set isolation level to READ COMMITTED
            conn_with_isolation = await conn.execution_options(isolation_level="READ COMMITTED")
            async with conn_with_isolation.begin():
                # Configure session factory with the connection
                sess_factory = async_sessionmaker(
                    bind=conn_with_isolation,
                    expire_on_commit=False,
                )
                session = sess_factory()
                yield session
                await session.commit()



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass

from ai.backend.common.dependencies import DependencyComposer, DependencyStack
from ai.backend.common.plugin.event import EventDispatcherPluginContext
from ai.backend.common.plugin.hook import HookPluginContext
from ai.backend.manager.plugin.network import NetworkPluginContext

from .base import PluginsInput
from .event_dispatcher import EventDispatcherPluginDependency
from .hook import HookPluginDependency
from .network import NetworkPluginDependency


@dataclass
class PluginsResources:
    """Container for all plugin context resources."""

    hook_plugin_ctx: HookPluginContext
    network_plugin_ctx: NetworkPluginContext
    event_dispatcher_plugin_ctx: EventDispatcherPluginContext


class PluginsComposer(DependencyComposer[PluginsInput, PluginsResources]):
    """Composes plugin context dependencies (network, hook, event dispatcher).

    Monitoring plugins (error_monitor, stats_monitor) are initialized separately
    in MonitoringComposer (after DomainComposer) because they require
    error_log_repository which is only available post-Domain stage.

    Initializes plugin contexts in the following order:
    1. NetworkPluginContext
    2. HookPluginContext (dispatches ACTIVATE_MANAGER)
    3. EventDispatcherPluginContext
    """

    @property
    def stage_name(self) -> str:
        return "plugins"

    @asynccontextmanager
    async def compose(
        self,
        stack: DependencyStack,
        setup_input: PluginsInput,
    ) -> AsyncIterator[PluginsResources]:
        """Compose plugin context dependencies in order.

        Args:
            stack: The dependency stack to use for composition
            setup_input: Plugins input containing etcd and config

        Yields:
            PluginsResources containing all plugin contexts
        """
        network_plugin_ctx = await stack.enter_dependency(
            NetworkPluginDependency(),
            setup_input,
        )

        hook_plugin_ctx = await stack.enter_dependency(
            HookPluginDependency(),
            setup_input,
        )

        event_dispatcher_plugin_ctx = await stack.enter_dependency(
            EventDispatcherPluginDependency(),
            setup_input,
        )

        yield PluginsResources(
            hook_plugin_ctx=hook_plugin_ctx,
            network_plugin_ctx=network_plugin_ctx,
            event_dispatcher_plugin_ctx=event_dispatcher_plugin_ctx,
        )



"""Get merged configuration action."""

from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, override

from ai.backend.common.data.permission.types import RBACElementType, ScopeType
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.permission.types import RBACElementRef
from ai.backend.manager.services.app_config.actions.base import (
    AppConfigScopeAction,
    AppConfigScopeActionResult,
)


@dataclass
class GetMergedAppConfigAction(AppConfigScopeAction):
    """Action to get merged app configuration for a user."""

    user_id: str

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET

    @override
    def scope_type(self) -> ScopeType:
        return ScopeType.USER

    @override
    def scope_id(self) -> str:
        return self.user_id

    @override
    def target_element(self) -> RBACElementRef:
        return RBACElementRef(RBACElementType.USER, self.user_id)


@dataclass
class GetMergedAppConfigActionResult(AppConfigScopeActionResult):
    """Result of get merged app configuration action."""

    user_id: str
    merged_config: Mapping[str, Any]

    @override
    def scope_type(self) -> ScopeType:
        return ScopeType.USER

    @override
    def scope_id(self) -> str:
        return self.user_id



"""Resource slot type module registrar."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.manager.api.rest.middleware.auth import superadmin_required
from ai.backend.manager.api.rest.routing import RouteRegistry

from .handler import ResourceSlotHandler

if TYPE_CHECKING:
    from ai.backend.manager.api.rest.types import RouteDeps


def register_resource_slot_routes(
    handler: ResourceSlotHandler, route_deps: RouteDeps
) -> RouteRegistry:
    """Build the resource-slot-types sub-application."""
    reg = RouteRegistry.create("resource-slot-types", route_deps.cors_options)

    reg.add(
        "POST",
        "/search",
        handler.search_resource_slot_types,
        middlewares=[
            superadmin_required,
            route_deps.all_status_mw,
        ],
    )
    reg.add(
        "GET",
        "/{slot_name}",
        handler.get_resource_slot_type,
        middlewares=[
            superadmin_required,
            route_deps.all_status_mw,
        ],
    )
    return reg



from __future__ import annotations

from functools import lru_cache

import strawberry
from strawberry import Info

from ai.backend.manager.api.gql.adapter import PaginationOptions, PaginationSpec
from ai.backend.manager.api.gql.base import encode_cursor
from ai.backend.manager.api.gql.session.types import (
    SessionV2ConnectionGQL,
    SessionV2EdgeGQL,
    SessionV2FilterGQL,
    SessionV2GQL,
    SessionV2OrderByGQL,
)
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.models.session import SessionRow
from ai.backend.manager.repositories.base import QueryCondition
from ai.backend.manager.repositories.scheduler.options import SessionConditions, SessionOrders
from ai.backend.manager.services.session.actions.search import SearchSessionsAction


@lru_cache(maxsize=1)
def _get_session_pagination_spec() -> PaginationSpec:
    return PaginationSpec(
        forward_order=SessionOrders.created_at(ascending=False),
        backward_order=SessionOrders.created_at(ascending=True),
        forward_condition_factory=SessionConditions.by_cursor_forward,
        backward_condition_factory=SessionConditions.by_cursor_backward,
        tiebreaker_order=SessionRow.id.asc(),
    )


async def fetch_sessions(
    info: Info[StrawberryGQLContext],
    filter: SessionV2FilterGQL | None = None,
    order_by: list[SessionV2OrderByGQL] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
    base_conditions: list[QueryCondition] | None = None,
) -> SessionV2ConnectionGQL:
    querier = info.context.gql_adapter.build_querier(
        PaginationOptions(
            first=first,
            after=after,
            last=last,
            before=before,
            limit=limit,
            offset=offset,
        ),
        pagination_spec=_get_session_pagination_spec(),
        filter=filter,
        order_by=order_by,
        base_conditions=base_conditions,
    )

    action_result = await info.context.processors.session.search_sessions.wait_for_complete(
        SearchSessionsAction(querier=querier)
    )

    nodes = [SessionV2GQL.from_data(session_data) for session_data in action_result.data]
    edges = [SessionV2EdgeGQL(node=node, cursor=encode_cursor(node.id)) for node in nodes]

    return SessionV2ConnectionGQL(
        edges=edges,
        page_info=strawberry.relay.PageInfo(
            has_next_page=action_result.has_next_page,
            has_previous_page=action_result.has_previous_page,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
        count=action_result.total_count,
    )



import uuid
from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import RBACElementType
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.permission.types import RBACElementRef
from ai.backend.manager.services.group.actions.base import (
    GroupSingleEntityAction,
    GroupSingleEntityActionResult,
)


@dataclass
class PurgeGroupAction(GroupSingleEntityAction):
    group_id: uuid.UUID

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.PURGE

    @override
    def target_entity_id(self) -> str:
        return str(self.group_id)

    @override
    def target_element(self) -> RBACElementRef:
        return RBACElementRef(RBACElementType.PROJECT, str(self.group_id))


@dataclass
class PurgeGroupActionResult(GroupSingleEntityActionResult):
    group_id: uuid.UUID

    @override
    def target_entity_id(self) -> str:
        return str(self.group_id)
